import copy
import datetime
import logging
import os
import time
from collections import namedtuple

import numpy as np
import pandas as pd

from ..utils import check_directory_exists_and_if_not_mkdir, logger
from .base_sampler import (
    MCMCSampler,
    SamplerError,
    _sampling_convenience_dump,
    signal_wrapper,
)

ConvergenceInputs = namedtuple(
    "ConvergenceInputs",
    [
        "autocorr_c",
        "autocorr_tol",
        "autocorr_tau",
        "gradient_tau",
        "gradient_mean_log_posterior",
        "Q_tol",
        "safety",
        "burn_in_nact",
        "burn_in_fixed_discard",
        "mean_logl_frac",
        "thin_by_nact",
        "nsamples",
        "ignore_keys_for_tau",
        "min_tau",
        "niterations_per_check",
    ],
)


class Ptemcee(MCMCSampler):
    """bilby wrapper ptemcee (https://github.com/willvousden/ptemcee)

    All positional and keyword arguments (i.e., the args and kwargs) passed to
    `run_sampler` will be propagated to `ptemcee.Sampler`, see
    documentation for that class for further help. Under Other Parameters, we
    list commonly used kwargs and the bilby defaults.

    Parameters
    ==========
    nsamples: int, (5000)
        The requested number of samples. Note, in cases where the
        autocorrelation parameter is difficult to measure, it is possible to
        end up with more than nsamples.
    burn_in_nact, thin_by_nact: int, (50, 1)
        The number of burn-in autocorrelation times to discard and the thin-by
        factor. Increasing burn_in_nact increases the time required for burn-in.
        Increasing thin_by_nact increases the time required to obtain nsamples.
    burn_in_fixed_discard: int (0)
        A fixed number of samples to discard for burn-in
    mean_logl_frac: float, (0.0.1)
        The maximum fractional change the mean log-likelihood to accept
    autocorr_tol: int, (50)
        The minimum number of autocorrelation times needed to trust the
        estimate of the autocorrelation time.
    autocorr_c: int, (5)
        The step size for the window search used by emcee.autocorr.integrated_time
    safety: int, (1)
        A multiplicative factor for the estimated autocorrelation. Useful for
        cases where non-convergence can be observed by eye but the automated
        tools are failing.
    autocorr_tau: int, (1)
        The number of autocorrelation times to use in assessing if the
        autocorrelation time is stable.
    gradient_tau: float, (0.1)
        The maximum (smoothed) local gradient of the ACT estimate to allow.
        This ensures the ACT estimate is stable before finishing sampling.
    gradient_mean_log_posterior: float, (0.1)
        The maximum (smoothed) local gradient of the logliklilhood to allow.
        This ensures the ACT estimate is stable before finishing sampling.
    Q_tol: float (1.01)
        The maximum between-chain to within-chain tolerance allowed (akin to
        the Gelman-Rubin statistic).
    min_tau: int, (1)
        A minimum tau (autocorrelation time) to accept.
    check_point_delta_t: float, (600)
        The period with which to checkpoint (in seconds).
    threads: int, (1)
        If threads > 1, a MultiPool object is setup and used.
    exit_code: int, (77)
        The code on which the sampler exits.
    store_walkers: bool (False)
        If true, store the unthinned, unburnt chains in the result. Note, this
        is not recommended for cases where tau is large.
    ignore_keys_for_tau: str
        A pattern used to ignore keys in estimating the autocorrelation time.
    pos0: str, list, np.ndarray, dict
        If a string, one of "prior" or "minimize". For "prior", the initial
        positions of the sampler are drawn from the sampler. If "minimize",
        a scipy.optimize step is applied to all parameters a number of times.
        The walkers are then initialized from the range of values obtained.
        If a list, for the keys in the list the optimization step is applied,
        otherwise the initial points are drawn from the prior.
        If a :code:`numpy` array the shape should be
        :code:`(ntemps, nwalkers, ndim)`.
        If a :code:`dict`, this should be a dictionary with keys matching the
        :code:`search_parameter_keys`. Each entry should be an array with
        shape :code:`(ntemps, nwalkers)`.

    niterations_per_check: int (5)
        The number of iteration steps to take before checking ACT. This
        effectively pre-thins the chains. Larger values reduce the per-eval
        timing due to improved efficiency. But, if it is made too large the
        pre-thinning may be overly aggressive effectively wasting compute-time.
        If you see tau=1, then niterations_per_check is likely too large.


    Other Parameters
    ================
    nwalkers: int, (200)
        The number of walkers
    nsteps: int, (100)
        The number of steps to take
    ntemps: int (10)
        The number of temperatures used by ptemcee
    Tmax: float
        The maximum temperature

    """

    # Arguments used by ptemcee
    default_kwargs = dict(
        ntemps=10,
        nwalkers=100,
        Tmax=None,
        betas=None,
        a=2.0,
        adaptation_lag=10000,
        adaptation_time=100,
        random=None,
        adapt=False,
        swap_ratios=False,
    )

    def __init__(
        self,
        likelihood,
        priors,
        outdir="outdir",
        label="label",
        use_ratio=False,
        check_point_plot=True,
        skip_import_verification=False,
        resume=True,
        nsamples=5000,
        burn_in_nact=50,
        burn_in_fixed_discard=0,
        mean_logl_frac=0.01,
        thin_by_nact=0.5,
        autocorr_tol=50,
        autocorr_c=5,
        safety=1,
        autocorr_tau=1,
        gradient_tau=0.1,
        gradient_mean_log_posterior=0.1,
        Q_tol=1.02,
        min_tau=1,
        check_point_delta_t=600,
        threads=1,
        exit_code=77,
        plot=False,
        store_walkers=False,
        ignore_keys_for_tau=None,
        pos0="prior",
        niterations_per_check=5,
        log10beta_min=None,
        verbose=True,
        **kwargs,
    ):
        super(Ptemcee, self).__init__(
            likelihood=likelihood,
            priors=priors,
            outdir=outdir,
            label=label,
            use_ratio=use_ratio,
            plot=plot,
            skip_import_verification=skip_import_verification,
            exit_code=exit_code,
            **kwargs,
        )

        self.nwalkers = self.sampler_init_kwargs["nwalkers"]
        self.ntemps = self.sampler_init_kwargs["ntemps"]
        self.max_steps = 500

        # Checkpointing inputs
        self.resume = resume
        self.check_point_delta_t = check_point_delta_t
        self.check_point_plot = check_point_plot
        self.resume_file = f"{self.outdir}/{self.label}_checkpoint_resume.pickle"

        # Store convergence checking inputs in a named tuple
        convergence_inputs_dict = dict(
            autocorr_c=autocorr_c,
            autocorr_tol=autocorr_tol,
            autocorr_tau=autocorr_tau,
            safety=safety,
            burn_in_nact=burn_in_nact,
            burn_in_fixed_discard=burn_in_fixed_discard,
            mean_logl_frac=mean_logl_frac,
            thin_by_nact=thin_by_nact,
            gradient_tau=gradient_tau,
            gradient_mean_log_posterior=gradient_mean_log_posterior,
            Q_tol=Q_tol,
            nsamples=nsamples,
            ignore_keys_for_tau=ignore_keys_for_tau,
            min_tau=min_tau,
            niterations_per_check=niterations_per_check,
        )
        self.convergence_inputs = ConvergenceInputs(**convergence_inputs_dict)
        logger.info(f"Using convergence inputs: {self.convergence_inputs}")

        # Check if threads was given as an equivalent arg
        if threads == 1:
            for equiv in self.npool_equiv_kwargs:
                if equiv in kwargs:
                    threads = kwargs.pop(equiv)

        # Store threads
        self.threads = threads

        # Misc inputs
        self.store_walkers = store_walkers
        self.pos0 = pos0

        self._periodic = [
            self.priors[key].boundary == "periodic"
            for key in self.search_parameter_keys
        ]
        self.priors.sample()
        self._minima = np.array(
            [self.priors[key].minimum for key in self.search_parameter_keys]
        )
        self._range = (
            np.array([self.priors[key].maximum for key in self.search_parameter_keys])
            - self._minima
        )

        self.log10beta_min = log10beta_min
        if self.log10beta_min is not None:
            betas = np.logspace(0, self.log10beta_min, self.ntemps)
            logger.warning(f"Using betas {betas}")
            self.kwargs["betas"] = betas
        self.verbose = verbose

        self.iteration = 0
        self.chain_array = self.get_zero_chain_array()
        self.log_likelihood_array = self.get_zero_array()
        self.log_posterior_array = self.get_zero_array()
        self.beta_list = list()
        self.tau_list = list()
        self.tau_list_n = list()
        self.Q_list = list()
        self.time_per_check = list()

        self.nburn = np.nan
        self.thin = np.nan
        self.tau_int = np.nan
        self.nsamples_effective = 0
        self.discard = 0

    @property
    def sampler_function_kwargs(self):
        """Kwargs passed to samper.sampler()"""
        keys = ["adapt", "swap_ratios"]
        return {key: self.kwargs[key] for key in keys}

    @property
    def sampler_init_kwargs(self):
        """Kwargs passed to initialize ptemcee.Sampler()"""
        return {
            key: value
            for key, value in self.kwargs.items()
            if key not in self.sampler_function_kwargs
        }

    def _translate_kwargs(self, kwargs):
        """Translate kwargs"""
        if "nwalkers" not in kwargs:
            for equiv in self.nwalkers_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["nwalkers"] = kwargs.pop(equiv)

    def get_pos0_from_prior(self):
        """Draw the initial positions from the prior

        Returns
        =======
        pos0: list
            The initial postitions of the walkers, with shape (ntemps, nwalkers, ndim)

        """
        logger.info("Generating pos0 samples")
        return np.array(
            [
                [self.get_random_draw_from_prior() for _ in range(self.nwalkers)]
                for _ in range(self.kwargs["ntemps"])
            ]
        )

    def get_pos0_from_minimize(self, minimize_list=None):
        """Draw the initial positions using an initial minimization step

        See pos0 in the class initialization for details.

        Returns
        =======
        pos0: list
            The initial postitions of the walkers, with shape (ntemps, nwalkers, ndim)

        """

        from scipy.optimize import minimize

        # Set up the minimize list: keys not in this list will have initial
        # positions drawn from the prior
        if minimize_list is None:
            minimize_list = self.search_parameter_keys
            pos0 = np.zeros((self.kwargs["ntemps"], self.kwargs["nwalkers"], self.ndim))
        else:
            pos0 = np.array(self.get_pos0_from_prior())

        logger.info(f"Attempting to set pos0 for {minimize_list} from minimize")

        likelihood_copy = copy.copy(self.likelihood)

        def neg_log_like(params):
            """Internal function to minimize"""
            likelihood_copy.parameters.update(
                {key: val for key, val in zip(minimize_list, params)}
            )
            try:
                return -likelihood_copy.log_likelihood()
            except RuntimeError:
                return +np.inf

        # Bounds used in the minimization
        bounds = [
            (self.priors[key].minimum, self.priors[key].maximum)
            for key in minimize_list
        ]

        # Run the minimization step several times to get a range of values
        trials = 0
        success = []
        while True:
            draw = self.priors.sample()
            likelihood_copy.parameters.update(draw)
            x0 = [draw[key] for key in minimize_list]
            res = minimize(
                neg_log_like, x0, bounds=bounds, method="L-BFGS-B", tol=1e-15
            )
            if res.success:
                success.append(res.x)
            if trials > 100:
                raise SamplerError("Unable to set pos0 from minimize")
            if len(success) >= 10:
                break

        # Initialize positions from the range of values
        success = np.array(success)
        for i, key in enumerate(minimize_list):
            pos0_min = np.min(success[:, i])
            pos0_max = np.max(success[:, i])
            logger.info(f"Initialize {key} walkers from {pos0_min}->{pos0_max}")
            j = self.search_parameter_keys.index(key)
            pos0[:, :, j] = np.random.uniform(
                pos0_min,
                pos0_max,
                size=(self.kwargs["ntemps"], self.kwargs["nwalkers"]),
            )
        return pos0

    def get_pos0_from_array(self):
        if self.pos0.shape != (self.ntemps, self.nwalkers, self.ndim):
            raise ValueError(
                "Shape of starting array should be (ntemps, nwalkers, ndim). "
                f"In this case that is ({self.ntemps}, {self.nwalkers}, "
                f"{self.ndim}), got {self.pos0.shape}"
            )
        else:
            return self.pos0

    def get_pos0_from_dict(self):
        """
        Initialize the starting points from a passed dictionary.

        The :code:`pos0` passed to the :code:`Sampler` should be a dictionary
        with keys matching the :code:`search_parameter_keys`.
        Each entry should have shape :code:`(ntemps, nwalkers)`.
        """
        pos0 = np.array([self.pos0[key] for key in self.search_parameter_keys])
        self.pos0 = np.moveaxis(pos0, 0, -1)
        return self.get_pos0_from_array()

    def setup_sampler(self):
        """Either initialize the sampler or read in the resume file"""
        import ptemcee

        if os.path.isfile(self.resume_file) and self.resume is True:
            import dill

            logger.info(f"Resume data {self.resume_file} found")
            with open(self.resume_file, "rb") as file:
                data = dill.load(file)

            # Extract the check-point data
            self.sampler = data["sampler"]
            self.iteration = data["iteration"]
            self.chain_array = data["chain_array"]
            self.log_likelihood_array = data["log_likelihood_array"]
            self.log_posterior_array = data["log_posterior_array"]
            self.pos0 = data["pos0"]
            self.beta_list = data["beta_list"]
            self.sampler._betas = np.array(self.beta_list[-1])
            self.tau_list = data["tau_list"]
            self.tau_list_n = data["tau_list_n"]
            self.Q_list = data["Q_list"]
            self.time_per_check = data["time_per_check"]

            # Initialize the pool
            self.sampler.pool = self.pool
            self.sampler.threads = self.threads

            logger.info(f"Resuming from previous run with time={self.iteration}")

        else:
            # Initialize the PTSampler
            if self.threads == 1:
                self.sampler = ptemcee.Sampler(
                    dim=self.ndim,
                    logl=self.log_likelihood,
                    logp=self.log_prior,
                    **self.sampler_init_kwargs,
                )
            else:
                self.sampler = ptemcee.Sampler(
                    dim=self.ndim,
                    logl=do_nothing_function,
                    logp=do_nothing_function,
                    threads=self.threads,
                    **self.sampler_init_kwargs,
                )

            self.sampler._likeprior = LikePriorEvaluator()

            # Initialize storing results
            self.iteration = 0
            self.chain_array = self.get_zero_chain_array()
            self.log_likelihood_array = self.get_zero_array()
            self.log_posterior_array = self.get_zero_array()
            self.beta_list = list()
            self.tau_list = list()
            self.tau_list_n = list()
            self.Q_list = list()
            self.time_per_check = list()
            self.pos0 = self.get_pos0()

        return self.sampler

    def get_zero_chain_array(self):
        return np.zeros((self.nwalkers, self.max_steps, self.ndim))

    def get_zero_array(self):
        return np.zeros((self.ntemps, self.nwalkers, self.max_steps))

    def get_pos0(self):
        """Master logic for setting pos0"""
        if isinstance(self.pos0, str) and self.pos0.lower() == "prior":
            return self.get_pos0_from_prior()
        elif isinstance(self.pos0, str) and self.pos0.lower() == "minimize":
            return self.get_pos0_from_minimize()
        elif isinstance(self.pos0, list):
            return self.get_pos0_from_minimize(minimize_list=self.pos0)
        elif isinstance(self.pos0, np.ndarray):
            return self.get_pos0_from_array()
        elif isinstance(self.pos0, dict):
            return self.get_pos0_from_dict()
        else:
            raise SamplerError(f"pos0={self.pos0} not implemented")

    def _close_pool(self):
        if getattr(self.sampler, "pool", None) is not None:
            self.sampler.pool = None
        if "pool" in self.result.sampler_kwargs:
            del self.result.sampler_kwargs["pool"]
        super(Ptemcee, self)._close_pool()

    @signal_wrapper
    def run_sampler(self):
        self._setup_pool()
        sampler = self.setup_sampler()

        t0 = datetime.datetime.now()
        logger.info("Starting to sample")
        while True:
            for (pos0, log_posterior, log_likelihood) in sampler.sample(
                self.pos0,
                storechain=False,
                iterations=self.convergence_inputs.niterations_per_check,
                **self.sampler_function_kwargs,
            ):
                pos0[:, :, self._periodic] = (
                    np.mod(
                        pos0[:, :, self._periodic] - self._minima[self._periodic],
                        self._range[self._periodic],
                    )
                    + self._minima[self._periodic]
                )

            if self.iteration == self.chain_array.shape[1]:
                self.chain_array = np.concatenate(
                    (self.chain_array, self.get_zero_chain_array()), axis=1
                )
                self.log_likelihood_array = np.concatenate(
                    (self.log_likelihood_array, self.get_zero_array()), axis=2
                )
                self.log_posterior_array = np.concatenate(
                    (self.log_posterior_array, self.get_zero_array()), axis=2
                )

            self.pos0 = pos0
            self.chain_array[:, self.iteration, :] = pos0[0, :, :]
            self.log_likelihood_array[:, :, self.iteration] = log_likelihood
            self.log_posterior_array[:, :, self.iteration] = log_posterior
            self.mean_log_posterior = np.mean(
                self.log_posterior_array[:, :, : self.iteration], axis=1
            )

            # Calculate time per iteration
            self.time_per_check.append((datetime.datetime.now() - t0).total_seconds())
            t0 = datetime.datetime.now()

            self.iteration += 1

            # Calculate minimum iteration step to discard
            minimum_iteration = get_minimum_stable_itertion(
                self.mean_log_posterior, frac=self.convergence_inputs.mean_logl_frac
            )
            logger.debug(f"Minimum iteration = {minimum_iteration}")

            # Calculate the maximum discard number
            discard_max = np.max(
                [self.convergence_inputs.burn_in_fixed_discard, minimum_iteration]
            )

            if self.iteration > discard_max + self.nwalkers:
                # If we have taken more than nwalkers steps after the discard
                # then set the discard
                self.discard = discard_max
            else:
                # If haven't discard everything (avoid initialisation bias)
                logger.debug("Too few steps to calculate convergence")
                self.discard = self.iteration

            (
                stop,
                self.nburn,
                self.thin,
                self.tau_int,
                self.nsamples_effective,
            ) = check_iteration(
                self.iteration,
                self.chain_array[:, self.discard : self.iteration, :],
                sampler,
                self.convergence_inputs,
                self.search_parameter_keys,
                self.time_per_check,
                self.beta_list,
                self.tau_list,
                self.tau_list_n,
                self.Q_list,
                self.mean_log_posterior,
                verbose=self.verbose,
            )

            if stop:
                logger.info("Finished sampling")
                break

            # If a checkpoint is due, checkpoint
            if os.path.isfile(self.resume_file):
                last_checkpoint_s = time.time() - os.path.getmtime(self.resume_file)
            else:
                last_checkpoint_s = np.sum(self.time_per_check)

            if last_checkpoint_s > self.check_point_delta_t:
                self.write_current_state(plot=self.check_point_plot)

        # Run a final checkpoint to update the plots and samples
        self.write_current_state(plot=self.check_point_plot)

        # Get 0-likelihood samples and store in the result
        self.result.samples = self.chain_array[
            :, self.discard + self.nburn : self.iteration : self.thin, :
        ].reshape((-1, self.ndim))
        loglikelihood = self.log_likelihood_array[
            0, :, self.discard + self.nburn : self.iteration : self.thin
        ]  # nwalkers, nsteps
        self.result.log_likelihood_evaluations = loglikelihood.reshape((-1))

        if self.store_walkers:
            self.result.walkers = self.sampler.chain
        self.result.nburn = self.nburn
        self.result.discard = self.discard

        log_evidence, log_evidence_err = compute_evidence(
            sampler,
            self.log_likelihood_array,
            self.outdir,
            self.label,
            self.discard,
            self.nburn,
            self.thin,
            self.iteration,
        )
        self.result.log_evidence = log_evidence
        self.result.log_evidence_err = log_evidence_err

        self.result.sampling_time = datetime.timedelta(
            seconds=np.sum(self.time_per_check)
        )

        self._close_pool()

        return self.result

    def write_current_state(self, plot=True):
        check_directory_exists_and_if_not_mkdir(self.outdir)
        checkpoint(
            self.iteration,
            self.outdir,
            self.label,
            self.nsamples_effective,
            self.sampler,
            self.discard,
            self.nburn,
            self.thin,
            self.search_parameter_keys,
            self.resume_file,
            self.log_likelihood_array,
            self.log_posterior_array,
            self.chain_array,
            self.pos0,
            self.beta_list,
            self.tau_list,
            self.tau_list_n,
            self.Q_list,
            self.time_per_check,
        )

        if plot:
            try:
                # Generate the walkers plot diagnostic
                plot_walkers(
                    self.chain_array[:, : self.iteration, :],
                    self.nburn,
                    self.thin,
                    self.search_parameter_keys,
                    self.outdir,
                    self.label,
                    self.discard,
                )
            except Exception as e:
                logger.info(f"Walkers plot failed with exception {e}")

            try:
                # Generate the tau plot diagnostic if DEBUG
                if logger.level < logging.INFO:
                    plot_tau(
                        self.tau_list_n,
                        self.tau_list,
                        self.search_parameter_keys,
                        self.outdir,
                        self.label,
                        self.tau_int,
                        self.convergence_inputs.autocorr_tau,
                    )
            except Exception as e:
                logger.info(f"tau plot failed with exception {e}")

            try:
                plot_mean_log_posterior(
                    self.mean_log_posterior,
                    self.outdir,
                    self.label,
                )
            except Exception as e:
                logger.info(f"mean_logl plot failed with exception {e}")


def get_minimum_stable_itertion(mean_array, frac, nsteps_min=10):
    nsteps = mean_array.shape[1]
    if nsteps < nsteps_min:
        return 0

    min_it = 0
    for x in mean_array:
        maxl = np.max(x)
        fracdiff = (maxl - x) / np.abs(maxl)
        idxs = fracdiff < frac
        if np.sum(idxs) > 0:
            min_it = np.max([min_it, np.min(np.arange(len(idxs))[idxs])])
    return min_it


def check_iteration(
    iteration,
    samples,
    sampler,
    convergence_inputs,
    search_parameter_keys,
    time_per_check,
    beta_list,
    tau_list,
    tau_list_n,
    Q_list,
    mean_log_posterior,
    verbose=True,
):
    """Per-iteration logic to calculate the convergence check

    Parameters
    ==========
    convergence_inputs: bilby.core.sampler.ptemcee.ConvergenceInputs
        A named tuple of the convergence checking inputs
    search_parameter_keys: list
        A list of the search parameter keys
    time_per_check, tau_list, tau_list_n: list
        Lists used for tracking the run
    verbose: bool
        Whether to print the output

    Returns
    =======
    stop: bool
        A boolean flag, True if the stopping criteria has been met
    burn: int
        The number of burn-in steps to discard
    thin: int
        The thin-by factor to apply
    tau_int: int
        The integer estimated ACT
    nsamples_effective: int
        The effective number of samples after burning and thinning
    """

    ci = convergence_inputs
    # Note: nsteps is the number of steps in the samples while iterations is
    # the current iteration number. So iteration > nsteps by the number of
    # of discards
    nwalkers, nsteps, ndim = samples.shape

    tau_array = calculate_tau_array(samples, search_parameter_keys, ci)

    # Maximum over parameters, mean over walkers
    tau = np.max(np.mean(tau_array, axis=0))

    # Apply multiplicitive safety factor
    tau = ci.safety * tau

    # Store for convergence checking and plotting
    beta_list.append(list(sampler.betas))
    tau_list.append(list(np.mean(tau_array, axis=0)))
    tau_list_n.append(iteration)

    Q = get_Q_convergence(samples)
    Q_list.append(Q)

    if np.isnan(tau) or np.isinf(tau):
        if verbose:
            print_progress(
                iteration,
                sampler,
                time_per_check,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                np.nan,
                False,
                convergence_inputs,
                Q,
            )
        return False, np.nan, np.nan, np.nan, np.nan

    # Convert to an integer
    tau_int = int(np.ceil(tau))

    # Calculate the effective number of samples available
    nburn = int(ci.burn_in_nact * tau_int)
    thin = int(np.max([1, ci.thin_by_nact * tau_int]))
    samples_per_check = nwalkers / thin
    nsamples_effective = int(nwalkers * (nsteps - nburn) / thin)

    # Calculate convergence boolean
    converged = Q < ci.Q_tol and ci.nsamples < nsamples_effective
    logger.debug(
        f"Convergence: Q<Q_tol={Q < ci.Q_tol}, "
        f"nsamples<nsamples_effective={ci.nsamples < nsamples_effective}"
    )

    GRAD_WINDOW_LENGTH = nwalkers + 1
    nsteps_to_check = ci.autocorr_tau * np.max([2 * GRAD_WINDOW_LENGTH, tau_int])
    lower_tau_index = np.max([0, len(tau_list) - nsteps_to_check])
    check_taus = np.array(tau_list[lower_tau_index:])
    if not np.any(np.isnan(check_taus)) and check_taus.shape[0] > GRAD_WINDOW_LENGTH:
        gradient_tau = get_max_gradient(check_taus, axis=0, window_length=11)

        if gradient_tau < ci.gradient_tau:
            logger.debug(
                f"tau usable as {gradient_tau} < gradient_tau={ci.gradient_tau}"
            )
            tau_usable = True
        else:
            logger.debug(
                f"tau not usable as {gradient_tau} > gradient_tau={ci.gradient_tau}"
            )
            tau_usable = False

        check_mean_log_posterior = mean_log_posterior[:, -nsteps_to_check:]
        gradient_mean_log_posterior = get_max_gradient(
            check_mean_log_posterior,
            axis=1,
            window_length=GRAD_WINDOW_LENGTH,
            smooth=True,
        )

        if gradient_mean_log_posterior < ci.gradient_mean_log_posterior:
            logger.debug(
                f"tau usable as {gradient_mean_log_posterior} < "
                f"gradient_mean_log_posterior={ci.gradient_mean_log_posterior}"
            )
            tau_usable *= True
        else:
            logger.debug(
                f"tau not usable as {gradient_mean_log_posterior} > "
                f"gradient_mean_log_posterior={ci.gradient_mean_log_posterior}"
            )
            tau_usable = False

    else:
        logger.debug("ACT is nan")
        gradient_tau = np.nan
        gradient_mean_log_posterior = np.nan
        tau_usable = False

    if nsteps < tau_int * ci.autocorr_tol:
        logger.debug("ACT less than autocorr_tol")
        tau_usable = False
    elif tau_int < ci.min_tau:
        logger.debug("ACT less than min_tau")
        tau_usable = False

    # Print an update on the progress
    if verbose:
        print_progress(
            iteration,
            sampler,
            time_per_check,
            nsamples_effective,
            samples_per_check,
            tau_int,
            gradient_tau,
            gradient_mean_log_posterior,
            tau_usable,
            convergence_inputs,
            Q,
        )
    stop = converged and tau_usable
    return stop, nburn, thin, tau_int, nsamples_effective


def get_max_gradient(x, axis=0, window_length=11, polyorder=2, smooth=False):
    from scipy.signal import savgol_filter

    if smooth:
        x = savgol_filter(x, axis=axis, window_length=window_length, polyorder=3)
    return np.max(
        savgol_filter(
            x, axis=axis, window_length=window_length, polyorder=polyorder, deriv=1
        )
    )


def get_Q_convergence(samples):
    nwalkers, nsteps, ndim = samples.shape
    if nsteps > 1:
        W = np.mean(np.var(samples, axis=1), axis=0)
        per_walker_mean = np.mean(samples, axis=1)
        mean = np.mean(per_walker_mean, axis=0)
        B = nsteps / (nwalkers - 1.0) * np.sum((per_walker_mean - mean) ** 2, axis=0)
        Vhat = (nsteps - 1) / nsteps * W + (nwalkers + 1) / (nwalkers * nsteps) * B
        Q_per_dim = np.sqrt(Vhat / W)
        return np.max(Q_per_dim)
    else:
        return np.inf


def print_progress(
    iteration,
    sampler,
    time_per_check,
    nsamples_effective,
    samples_per_check,
    tau_int,
    gradient_tau,
    gradient_mean_log_posterior,
    tau_usable,
    convergence_inputs,
    Q,
):
    # Setup acceptance string
    acceptance = sampler.acceptance_fraction[0, :]
    acceptance_str = f"{np.min(acceptance):1.2f}-{np.max(acceptance):1.2f}"

    # Setup tswap acceptance string
    tswap_acceptance_fraction = sampler.tswap_acceptance_fraction
    tswap_acceptance_str = f"{np.min(tswap_acceptance_fraction):1.2f}-{np.max(tswap_acceptance_fraction):1.2f}"

    ave_time_per_check = np.mean(time_per_check[-3:])
    time_left = (
        (convergence_inputs.nsamples - nsamples_effective)
        * ave_time_per_check
        / samples_per_check
    )
    if time_left > 0:
        time_left = str(datetime.timedelta(seconds=int(time_left)))
    else:
        time_left = "waiting on convergence"

    sampling_time = datetime.timedelta(seconds=np.sum(time_per_check))

    tau_str = f"{tau_int}(+{gradient_tau:0.2f},+{gradient_mean_log_posterior:0.2f})"

    if tau_usable:
        tau_str = f"={tau_str}"
    else:
        tau_str = f"!{tau_str}"

    Q_str = f"{Q:0.2f}"

    evals_per_check = (
        sampler.nwalkers * sampler.ntemps * convergence_inputs.niterations_per_check
    )

    approximate_ncalls = (
        convergence_inputs.niterations_per_check
        * iteration
        * sampler.nwalkers
        * sampler.ntemps
    )
    ncalls = f"{approximate_ncalls:1.1e}"
    eval_timing = f"{1000.0 * ave_time_per_check / evals_per_check:1.2f}ms/ev"

    try:
        print(
            f"{iteration}|{str(sampling_time).split('.')[0]}|nc:{ncalls}|"
            f"a0:{acceptance_str}|swp:{tswap_acceptance_str}|"
            f"n:{nsamples_effective}<{convergence_inputs.nsamples}|t{tau_str}|"
            f"q:{Q_str}|{eval_timing}",
            flush=True,
        )
    except OSError as e:
        logger.debug(f"Failed to print iteration due to :{e}")


def calculate_tau_array(samples, search_parameter_keys, ci):
    """Compute ACT tau for 0-temperature chains"""
    import emcee

    nwalkers, nsteps, ndim = samples.shape
    tau_array = np.zeros((nwalkers, ndim)) + np.inf
    if nsteps > 1:
        for ii in range(nwalkers):
            for jj, key in enumerate(search_parameter_keys):
                if ci.ignore_keys_for_tau and ci.ignore_keys_for_tau in key:
                    continue
                try:
                    tau_array[ii, jj] = emcee.autocorr.integrated_time(
                        samples[ii, :, jj], c=ci.autocorr_c, tol=0
                    )[0]
                except emcee.autocorr.AutocorrError:
                    tau_array[ii, jj] = np.inf
    return tau_array


def checkpoint(
    iteration,
    outdir,
    label,
    nsamples_effective,
    sampler,
    discard,
    nburn,
    thin,
    search_parameter_keys,
    resume_file,
    log_likelihood_array,
    log_posterior_array,
    chain_array,
    pos0,
    beta_list,
    tau_list,
    tau_list_n,
    Q_list,
    time_per_check,
):
    import dill

    logger.info("Writing checkpoint and diagnostics")
    ndim = sampler.dim

    # Store the samples if possible
    if nsamples_effective > 0:
        filename = f"{outdir}/{label}_samples.txt"
        samples = np.array(chain_array)[
            :, discard + nburn : iteration : thin, :
        ].reshape((-1, ndim))
        df = pd.DataFrame(samples, columns=search_parameter_keys)
        df.to_csv(filename, index=False, header=True, sep=" ")

    # Pickle the resume artefacts
    pool = sampler.pool
    sampler.pool = None
    sampler_copy = copy.deepcopy(sampler)
    sampler.pool = pool

    data = dict(
        iteration=iteration,
        sampler=sampler_copy,
        beta_list=beta_list,
        tau_list=tau_list,
        tau_list_n=tau_list_n,
        Q_list=Q_list,
        time_per_check=time_per_check,
        log_likelihood_array=log_likelihood_array,
        log_posterior_array=log_posterior_array,
        chain_array=chain_array,
        pos0=pos0,
    )

    with open(resume_file, "wb") as file:
        dill.dump(data, file, protocol=4)
    del data, sampler_copy
    logger.info("Finished writing checkpoint")


def plot_walkers(walkers, nburn, thin, parameter_labels, outdir, label, discard=0):
    """Method to plot the trace of the walkers in an ensemble MCMC plot"""
    import matplotlib.pyplot as plt

    nwalkers, nsteps, ndim = walkers.shape
    if np.isnan(nburn):
        nburn = nsteps
    if np.isnan(thin):
        thin = 1
    idxs = np.arange(nsteps)
    fig, axes = plt.subplots(nrows=ndim, ncols=2, figsize=(8, 3 * ndim))
    scatter_kwargs = dict(
        lw=0,
        marker="o",
        markersize=1,
        alpha=0.1,
    )

    # Plot the fixed burn-in
    if discard > 0:
        for i, (ax, axh) in enumerate(axes):
            ax.plot(
                idxs[:discard],
                walkers[:, :discard, i].T,
                color="gray",
                **scatter_kwargs,
            )

    # Plot the burn-in
    for i, (ax, axh) in enumerate(axes):
        ax.plot(
            idxs[discard : discard + nburn + 1],
            walkers[:, discard : discard + nburn + 1, i].T,
            color="C1",
            **scatter_kwargs,
        )

    # Plot the thinned posterior samples
    for i, (ax, axh) in enumerate(axes):
        ax.plot(
            idxs[discard + nburn :: thin],
            walkers[:, discard + nburn :: thin, i].T,
            color="C0",
            **scatter_kwargs,
        )
        axh.hist(
            walkers[:, discard + nburn :: thin, i].reshape((-1)), bins=50, alpha=0.8
        )

    for i, (ax, axh) in enumerate(axes):
        axh.set_xlabel(parameter_labels[i])
        ax.set_ylabel(parameter_labels[i])

    fig.tight_layout()
    filename = f"{outdir}/{label}_checkpoint_trace.png"
    fig.savefig(filename)
    plt.close(fig)


def plot_tau(
    tau_list_n,
    tau_list,
    search_parameter_keys,
    outdir,
    label,
    tau,
    autocorr_tau,
):
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()
    for i, key in enumerate(search_parameter_keys):
        ax.plot(tau_list_n, np.array(tau_list)[:, i], label=key)
    ax.set_xlabel("Iteration")
    ax.set_ylabel(r"$\langle \tau \rangle$")
    ax.legend()
    fig.tight_layout()
    fig.savefig(f"{outdir}/{label}_checkpoint_tau.png")
    plt.close(fig)


def plot_mean_log_posterior(mean_log_posterior, outdir, label):
    import matplotlib.pyplot as plt

    ntemps, nsteps = mean_log_posterior.shape
    ymax = np.max(mean_log_posterior)
    ymin = np.min(mean_log_posterior[:, -100:])
    ymax += 0.1 * (ymax - ymin)
    ymin -= 0.1 * (ymax - ymin)

    fig, ax = plt.subplots()
    idxs = np.arange(nsteps)
    ax.plot(idxs, mean_log_posterior.T)
    ax.set(
        xlabel="Iteration",
        ylabel=r"$\langle\mathrm{log-posterior}\rangle$",
        ylim=(ymin, ymax),
    )
    fig.tight_layout()
    fig.savefig(f"{outdir}/{label}_checkpoint_meanlogposterior.png")
    plt.close(fig)


def compute_evidence(
    sampler,
    log_likelihood_array,
    outdir,
    label,
    discard,
    nburn,
    thin,
    iteration,
    make_plots=True,
):
    """Computes the evidence using thermodynamic integration"""
    import matplotlib.pyplot as plt

    betas = sampler.betas
    # We compute the evidence without the burnin samples, but we do not thin
    lnlike = log_likelihood_array[:, :, discard + nburn : iteration]
    mean_lnlikes = np.mean(np.mean(lnlike, axis=1), axis=1)

    mean_lnlikes = mean_lnlikes[::-1]
    betas = betas[::-1]

    if any(np.isinf(mean_lnlikes)):
        logger.warning(
            "mean_lnlikes contains inf: recalculating without"
            f" the {len(betas[np.isinf(mean_lnlikes)])} infs"
        )
        idxs = np.isinf(mean_lnlikes)
        mean_lnlikes = mean_lnlikes[~idxs]
        betas = betas[~idxs]

    lnZ = np.trapz(mean_lnlikes, betas)
    z1 = np.trapz(mean_lnlikes, betas)
    z2 = np.trapz(mean_lnlikes[::-1][::2][::-1], betas[::-1][::2][::-1])
    lnZerr = np.abs(z1 - z2)

    if make_plots:
        fig, (ax1, ax2) = plt.subplots(nrows=2, figsize=(6, 8))
        ax1.semilogx(betas, mean_lnlikes, "-o")
        ax1.set_xlabel(r"$\beta$")
        ax1.set_ylabel(r"$\langle \log(\mathcal{L}) \rangle$")
        min_betas = []
        evidence = []
        for i in range(int(len(betas) / 2.0)):
            min_betas.append(betas[i])
            evidence.append(np.trapz(mean_lnlikes[i:], betas[i:]))

        ax2.semilogx(min_betas, evidence, "-o")
        ax2.set_ylabel(
            r"$\int_{\beta_{min}}^{\beta=1}"
            + r"\langle \log(\mathcal{L})\rangle d\beta$",
            size=16,
        )
        ax2.set_xlabel(r"$\beta_{min}$")
        plt.tight_layout()
        fig.savefig(f"{outdir}/{label}_beta_lnl.png")
        plt.close(fig)

    return lnZ, lnZerr


def do_nothing_function():
    """This is a do-nothing function, we overwrite the likelihood and prior elsewhere"""
    pass


class LikePriorEvaluator(object):
    """
    This class is copied and modified from ptemcee.LikePriorEvaluator, see
    https://github.com/willvousden/ptemcee for the original version

    We overwrite the logl and logp methods in order to improve the performance
    when using a MultiPool object: essentially reducing the amount of data
    transfer overhead.

    """

    def __init__(self):
        self.periodic_set = False

    def _setup_periodic(self):
        priors = _sampling_convenience_dump.priors
        search_parameter_keys = _sampling_convenience_dump.search_parameter_keys
        self._periodic = [
            priors[key].boundary == "periodic" for key in search_parameter_keys
        ]
        priors.sample()
        self._minima = np.array([priors[key].minimum for key in search_parameter_keys])
        self._range = (
            np.array([priors[key].maximum for key in search_parameter_keys])
            - self._minima
        )
        self.periodic_set = True

    def _wrap_periodic(self, array):
        if not self.periodic_set:
            self._setup_periodic()
        array[self._periodic] = (
            np.mod(
                array[self._periodic] - self._minima[self._periodic],
                self._range[self._periodic],
            )
            + self._minima[self._periodic]
        )
        return array

    def logl(self, v_array):
        priors = _sampling_convenience_dump.priors
        likelihood = _sampling_convenience_dump.likelihood
        search_parameter_keys = _sampling_convenience_dump.search_parameter_keys
        parameters = {key: v for key, v in zip(search_parameter_keys, v_array)}
        if priors.evaluate_constraints(parameters) > 0:
            likelihood.parameters.update(parameters)
            if _sampling_convenience_dump.use_ratio:
                return likelihood.log_likelihood() - likelihood.noise_log_likelihood()
            else:
                return likelihood.log_likelihood()
        else:
            return np.nan_to_num(-np.inf)

    def logp(self, v_array):
        priors = _sampling_convenience_dump.priors
        search_parameter_keys = _sampling_convenience_dump.search_parameter_keys
        params = {key: t for key, t in zip(search_parameter_keys, v_array)}
        return priors.ln_prob(params)

    def call_emcee(self, theta):
        ll, lp = self.__call__(theta)
        return ll + lp, [ll, lp]

    def __call__(self, x):
        lp = self.logp(x)
        if np.isnan(lp):
            raise ValueError("Prior function returned NaN.")

        if lp == float("-inf"):
            # Can't return -inf, since this messes with beta=0 behaviour.
            ll = 0
        else:
            ll = self.logl(x)
            if np.isnan(ll).any():
                raise ValueError("Log likelihood function returned NaN.")

        return ll, lp



from distutils.version import LooseVersion

import numpy as np
import pandas as pd

from ..core.sampler.base_sampler import SamplerError
from ..core.utils import logger
from .utils import LOGLKEY, LOGLLATEXKEY, LOGPKEY, LOGPLATEXKEY


class Chain(object):
    def __init__(
        self,
        initial_sample,
        burn_in_nact=1,
        thin_by_nact=1,
        fixed_discard=0,
        autocorr_c=5,
        min_tau=1,
        fixed_tau=None,
        tau_window=None,
        block_length=100000,
    ):
        """Object to store a single mcmc chain

        Parameters
        ----------
        initial_sample: bilby.bilby_mcmc.chain.Sample
            The starting point of the chain
        burn_in_nact, thin_by_nact : int (1, 1)
            The number of autocorrelation times (tau) to discard for burn-in
            and the multiplicative factor to thin by (thin_by_nact < 1). I.e
            burn_in_nact=10 and thin_by_nact=1 will discard 10*tau samples from
            the start of the chain, then thin the final chain by a factor
            of 1*tau (resulting in independent samples).
        fixed_discard: int (0)
            A fixed minimum number of samples to discard (can be used to
            override the burn_in_nact if it is too small).
        autocorr_c: float (5)
            The step size of the window search used by emcee.autocorr when
            estimating the autocorrelation time.
        min_tau: int (1)
            A minimum value for the autocorrelation time.
        fixed_tau: int (None)
            A fixed value for the autocorrelation (overrides the automated
            autocorrelation time estimation). Used in testing.
        tau_window: int (None)
            Only calculate the autocorrelation time in a trailing window. If
            None (default) this method is not used.
        block_length: int
            The incremental size to extend the array by when it runs out of
            space.
        """
        self.autocorr_c = autocorr_c
        self.min_tau = min_tau
        self.burn_in_nact = burn_in_nact
        self.thin_by_nact = thin_by_nact
        self.block_length = block_length
        self.fixed_discard = int(fixed_discard)
        self.fixed_tau = fixed_tau
        self.tau_window = tau_window

        self.ndim = initial_sample.ndim
        self.current_sample = initial_sample
        self.keys = self.current_sample.keys
        self.parameter_keys = self.current_sample.parameter_keys

        # Initialize chain
        self._chain_array = self._get_zero_chain_array()
        self._chain_array_length = block_length
        self.position = -1
        self.max_log_likelihood = -np.inf
        self.max_tau_dict = {}
        self.converged = False
        self.cached_tau_count = 0
        self._minimum_index_proposal = 0
        self._minimum_index_adapt = 0
        self._last_minimum_index = (0, 0, "I")
        self.last_full_tau_dict = {key: np.inf for key in self.parameter_keys}

        # Append the initial sample
        self.append(self.current_sample)

    def _get_zero_chain_array(self):
        return np.zeros((self.block_length, self.ndim + 2), dtype=np.float64)

    def _extend_chain_array(self):
        self._chain_array = np.concatenate(
            (self._chain_array, self._get_zero_chain_array()), axis=0
        )
        self._chain_array_length = len(self._chain_array)

    @property
    def current_sample(self):
        return self._current_sample.copy()

    @current_sample.setter
    def current_sample(self, current_sample):
        self._current_sample = current_sample

    def append(self, sample):
        self.position += 1

        # Extend the array if needed
        if self.position >= self._chain_array_length:
            self._extend_chain_array()

        # Store the current sample and append to the array
        self.current_sample = sample
        self._chain_array[self.position] = sample.list

        # Update the maximum log_likelihood
        if sample[LOGLKEY] > self.max_log_likelihood:
            self.max_log_likelihood = sample[LOGLKEY]

    def __getitem__(self, index):
        if index < 0:
            index = index + self.position + 1

        if index <= self.position:
            values = self._chain_array[index]
            return Sample({k: v for k, v in zip(self.keys, values)})
        else:
            raise SamplerError(f"Requested index {index} out of bounds")

    def __setitem__(self, index, sample):
        if index < 0:
            index = index + self.position + 1

        self._chain_array[index] = sample.list

    def key_to_idx(self, key):
        return self.keys.index(key)

    def get_1d_array(self, key):
        return self._chain_array[: 1 + self.position, self.key_to_idx(key)]

    @property
    def _random_idx(self):
        mindex = self._last_minimum_index[1]
        # Check if mindex exceeds current position by 10 ACT: if so use a random sample
        # otherwise we draw only from the chain past the minimum_index
        if np.isinf(self.tau_last) or self.position - mindex < 10 * self.tau_last:
            mindex = 0
        return np.random.randint(mindex, self.position + 1)

    @property
    def random_sample(self):
        return self[self._random_idx]

    @property
    def fixed_discard(self):
        return self._fixed_discard

    @fixed_discard.setter
    def fixed_discard(self, fixed_discard):
        self._fixed_discard = int(fixed_discard)

    @property
    def minimum_index(self):
        """This calculated a minimum index from which to discard samples

        A number of methods are provided for the calculation. A subset are
        switched off (by `if False` statements) for future development

        """
        position = self.position

        # Return cached minimum index
        last_minimum_index = self._last_minimum_index
        if position == last_minimum_index[0]:
            return int(last_minimum_index[1])

        # If fixed discard is not yet reached, just return that
        if position < self.fixed_discard:
            self.minimum_index_method = "FD"
            return self.fixed_discard

        # Initialize list of minimum index methods with the fixed discard (FD)
        minimum_index_list = [self.fixed_discard]
        minimum_index_method_list = ["FD"]

        # Calculate minimum index from tau
        if self.tau_last < np.inf:
            tau = self.tau_last
        elif len(self.max_tau_dict) == 0:
            # Bootstrap calculating tau when minimum index has not yet been calculated
            tau = self._tau_for_full_chain
        else:
            tau = np.inf

        if tau < np.inf:
            minimum_index_list.append(self.burn_in_nact * tau)
            minimum_index_method_list.append(f"{self.burn_in_nact}tau")

        # Calculate points when log-posterior is within z std of the mean
        if True:
            zfactor = 1
            N = 100
            delta_lnP = zfactor * self.ndim / 2
            logl = self.get_1d_array(LOGLKEY)
            log_prior = self.get_1d_array(LOGPKEY)
            log_posterior = logl + log_prior
            max_posterior = np.max(log_posterior)

            ave = pd.Series(log_posterior).rolling(window=N).mean().iloc[N - 1 :]
            delta = max_posterior - ave
            passes = ave[delta < delta_lnP]
            if len(passes) > 0:
                minimum_index_list.append(passes.index[0] + 1)
                minimum_index_method_list.append(f"z{zfactor}")

        # Add last minimum_index_method
        if False:
            minimum_index_list.append(last_minimum_index[1])
            minimum_index_method_list.append(last_minimum_index[2])

        # Minimum index set by proposals
        minimum_index_list.append(self.minimum_index_proposal)
        minimum_index_method_list.append("PR")

        # Minimum index set by temperature adaptation
        minimum_index_list.append(self.minimum_index_adapt)
        minimum_index_method_list.append("AD")

        # Calculate the maximum minimum index and associated method (reporting)
        minimum_index = int(np.max(minimum_index_list))
        minimum_index_method = minimum_index_method_list[np.argmax(minimum_index_list)]

        # Cache the method
        self._last_minimum_index = (position, minimum_index, minimum_index_method)
        self.minimum_index_method = minimum_index_method

        return minimum_index

    @property
    def minimum_index_proposal(self):
        return self._minimum_index_proposal

    @minimum_index_proposal.setter
    def minimum_index_proposal(self, minimum_index_proposal):
        if minimum_index_proposal > self._minimum_index_proposal:
            self._minimum_index_proposal = minimum_index_proposal

    @property
    def minimum_index_adapt(self):
        return self._minimum_index_adapt

    @minimum_index_adapt.setter
    def minimum_index_adapt(self, minimum_index_adapt):
        if minimum_index_adapt > self._minimum_index_adapt:
            self._minimum_index_adapt = minimum_index_adapt

    @property
    def tau(self):
        """The maximum ACT over all parameters"""

        if self.position in self.max_tau_dict:
            # If we have the ACT at the current position, return it
            return self.max_tau_dict[self.position]
        elif (
            self.tau_last < np.inf
            and self.cached_tau_count < 50
            and self.nsamples_last > 50
        ):
            # If we have a recent ACT return it
            self.cached_tau_count += 1
            return self.tau_last
        else:
            # Calculate the ACT
            return self.tau_nocache

    @property
    def tau_nocache(self):
        """Calculate tau forcing a recalculation (no cached tau)"""
        tau = max(self.tau_dict.values())
        self.max_tau_dict[self.position] = tau
        self.cached_tau_count = 0
        return tau

    @property
    def tau_last(self):
        """Return the last-calculated tau if it exists, else inf"""
        if len(self.max_tau_dict) > 0:
            return list(self.max_tau_dict.values())[-1]
        else:
            return np.inf

    @property
    def _tau_for_full_chain(self):
        """The maximum ACT over all parameters"""
        return max(self._tau_dict_for_full_chain.values())

    @property
    def _tau_dict_for_full_chain(self):
        return self._calculate_tau_dict(minimum_index=0)

    @property
    def tau_dict(self):
        """Calculate a dictionary of tau (ACT) for every parameter"""
        return self._calculate_tau_dict(self.minimum_index)

    def _calculate_tau_dict(self, minimum_index):
        """Calculate a dictionary of tau (ACT) for every parameter"""
        logger.debug(f"Calculating tau_dict {self}")

        # If there are too few samples to calculate tau
        if (self.position - minimum_index) < 2 * self.autocorr_c:
            return {key: np.inf for key in self.parameter_keys}

        # Choose minimimum index for the ACT calculation
        last_tau = self.tau_last
        if self.tau_window is not None and last_tau < np.inf:
            minimum_index_for_act = max(
                minimum_index, int(self.position - self.tau_window * last_tau)
            )
        else:
            minimum_index_for_act = minimum_index

        # Calculate a dictionary of tau's for each parameter
        taus = {}
        for key in self.parameter_keys:
            if self.fixed_tau is None:
                x = self.get_1d_array(key)[minimum_index_for_act:]
                tau = calculate_tau(x, self.autocorr_c)
                taux = round(tau, 1)
            else:
                taux = self.fixed_tau
            taus[key] = max(taux, self.min_tau)

        # Cache the last tau dictionary for future use
        self.last_full_tau_dict = taus

        return taus

    @property
    def thin(self):
        if np.isfinite(self.tau):
            return np.max([1, int(self.thin_by_nact * self.tau)])
        else:
            return 1

    @property
    def nsamples(self):
        nuseable_steps = self.position - self.minimum_index
        return int(nuseable_steps / (self.thin_by_nact * self.tau))

    @property
    def nsamples_last(self):
        nuseable_steps = self.position - self.minimum_index
        return int(nuseable_steps / (self.thin_by_nact * self.tau_last))

    @property
    def samples(self):
        samples = self._chain_array[self.minimum_index : self.position : self.thin]
        return pd.DataFrame(samples, columns=self.keys)

    def plot(self, outdir=".", label="label", priors=None, all_samples=None):
        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(
            nrows=self.ndim + 3, ncols=2, figsize=(8, 9 + 3 * (self.ndim))
        )
        scatter_kwargs = dict(
            lw=0,
            marker="o",
        )
        K = 1000

        nburn = self.minimum_index
        plot_setups = zip(
            [0, nburn, nburn],
            [nburn, self.position, self.position],
            [1, 1, self.thin],  # Thin-by factor
            ["tab:red", "tab:grey", "tab:blue"],  # Color
            [0.5, 0.05, 0.5],  # Alpha
            [1, 1, 1],  # Marker size
        )

        position_indexes = np.arange(self.position + 1)

        # Plot the traceplots
        for (start, stop, thin, color, alpha, ms) in plot_setups:
            for ax, key in zip(axes[:, 0], self.keys):
                xx = position_indexes[start:stop:thin] / K
                yy = self.get_1d_array(key)[start:stop:thin]

                # Downsample plots to max_pts: avoid memory issues
                max_pts = 10000
                while len(xx) > max_pts:
                    xx = xx[::2]
                    yy = yy[::2]

                ax.plot(
                    xx,
                    yy,
                    color=color,
                    alpha=alpha,
                    ms=ms,
                    **scatter_kwargs,
                )
                ax.set_ylabel(self._get_plot_label_by_key(key, priors))
                if key not in [LOGLKEY, LOGPKEY]:
                    msg = r"$\tau=$" + f"{self.last_full_tau_dict[key]:0.1f}"
                    ax.set_title(msg)

        # Plot the histograms
        for ax, key in zip(axes[:, 1], self.keys):
            if all_samples is not None:
                yy_all = all_samples[key]
                ax.hist(yy_all, bins=50, alpha=0.6, density=True, color="k")

            yy = self.get_1d_array(key)[nburn : self.position : self.thin]
            ax.hist(yy, bins=50, alpha=0.8, density=True)

            ax.set_xlabel(self._get_plot_label_by_key(key, priors))

        # Add x-axes labels to the traceplots
        axes[-1, 0].set_xlabel(r"Iteration $[\times 10^{3}]$")

        # Plot the calculated ACT
        ax = axes[-1, 0]
        tausit = np.array(list(self.max_tau_dict.keys()) + [self.position]) / K
        taus = list(self.max_tau_dict.values()) + [self.tau_last]
        ax.plot(tausit, taus, color="C3")
        ax.set(ylabel=r"Maximum $\tau$")

        axes[-1, 1].set_axis_off()

        filename = "{}/{}_checkpoint_trace.png".format(outdir, label)
        msg = [
            r"Maximum $\tau$" + f"={self.tau:0.1f} ",
            r"$n_{\rm samples}=$" + f"{self.nsamples} ",
        ]
        if self.thin_by_nact != 1:
            msg += [
                r"$n_{\rm samples}^{\rm eff}=$"
                + f"{int(self.nsamples * self.thin_by_nact)} "
            ]
        fig.suptitle(
            "| ".join(msg),
            y=1,
        )
        fig.tight_layout()
        fig.savefig(filename, dpi=200)
        plt.close(fig)

    @staticmethod
    def _get_plot_label_by_key(key, priors=None):
        if priors is not None and key in priors:
            return priors[key].latex_label
        elif key == LOGLKEY:
            return LOGLLATEXKEY
        elif key == LOGPKEY:
            return LOGPLATEXKEY
        else:
            return key


class Sample(object):
    def __init__(self, sample_dict):
        """A single sample

        Parameters
        ----------
        sample_dict: dict
            A dictionary of the sample
        """

        self.sample_dict = sample_dict
        self.keys = list(sample_dict.keys())
        self.parameter_keys = [k for k in self.keys if k not in [LOGPKEY, LOGLKEY]]
        self.ndim = len(self.parameter_keys)

    def __getitem__(self, key):
        return self.sample_dict[key]

    def __setitem__(self, key, value):
        self.sample_dict[key] = value
        if key not in self.keys:
            self.keys = list(self.sample_dict.keys())

    @property
    def list(self):
        return list(self.sample_dict.values())

    def __repr__(self):
        return str(self.sample_dict)

    @property
    def parameter_only_dict(self):
        return {key: self.sample_dict[key] for key in self.parameter_keys}

    @property
    def dict(self):
        return {key: self.sample_dict[key] for key in self.keys}

    def as_dict(self, keys=None):
        sdict = self.dict
        if keys is None:
            return sdict
        else:
            return {key: sdict[key] for key in keys}

    def __eq__(self, other_sample):
        return self.list == other_sample.list

    def copy(self):
        return Sample(self.sample_dict.copy())


def calculate_tau(x, autocorr_c=5):
    import emcee

    if LooseVersion(emcee.__version__) < LooseVersion("3"):
        raise SamplerError("bilby-mcmc requires emcee > 3.0 for autocorr analysis")

    if np.all(np.diff(x) == 0):
        return np.inf
    try:
        # Hard code tol=1: we perform this check internally
        tau = emcee.autocorr.integrated_time(x, c=autocorr_c, tol=1)[0]
        if np.isnan(tau):
            tau = np.inf
        return tau
    except emcee.autocorr.AutocorrError:
        return np.inf



import numpy as np

from ..result import read_in_result
from .base_sampler import Sampler


class FakeSampler(Sampler):
    """
    A "fake" sampler that evaluates the likelihood at a list of
    configurations read from a posterior data file.

    See base class for parameters. Added parameters are described below.

    Parameters
    ==========
    sample_file: str
        A string pointing to the posterior data file to be loaded.
    """

    default_kwargs = dict(
        verbose=True, logl_args=None, logl_kwargs=None, print_progress=True
    )

    def __init__(
        self,
        likelihood,
        priors,
        sample_file,
        outdir="outdir",
        label="label",
        use_ratio=False,
        plot=False,
        injection_parameters=None,
        meta_data=None,
        result_class=None,
        **kwargs
    ):
        super(FakeSampler, self).__init__(
            likelihood=likelihood,
            priors=priors,
            outdir=outdir,
            label=label,
            use_ratio=False,
            plot=False,
            skip_import_verification=True,
            injection_parameters=None,
            meta_data=None,
            result_class=None,
            **kwargs
        )
        self._read_parameter_list_from_file(sample_file)
        self.result.outdir = outdir
        self.result.label = label

    def _read_parameter_list_from_file(self, sample_file):
        """Read a list of sampling parameters from file.

        The sample_file should be in bilby posterior HDF5 format.
        """
        self.result = read_in_result(filename=sample_file)

    def run_sampler(self):
        """Compute the likelihood for the list of parameter space points."""
        self.sampler = "fake_sampler"

        # Flushes the output to force a line break
        if self.kwargs["verbose"]:
            print("")

        likelihood_ratios = []
        posterior = self.result.posterior

        for i in np.arange(posterior.shape[0]):
            sample = posterior.iloc[i]

            self.likelihood.parameters = sample.to_dict()
            logl = self.likelihood.log_likelihood_ratio()
            sample.log_likelihood = logl
            likelihood_ratios.append(logl)

            if self.kwargs["verbose"]:
                print(
                    self.likelihood.parameters["log_likelihood"],
                    likelihood_ratios[-1],
                    self.likelihood.parameters["log_likelihood"]
                    - likelihood_ratios[-1],
                )

        self.result.log_likelihood_evaluations = np.array(likelihood_ratios)

        self.result.log_evidence = np.nan
        self.result.log_evidence_err = np.nan

        return self.result



import os
import shutil
from collections import namedtuple
from distutils.version import LooseVersion
from shutil import copyfile

import numpy as np
from pandas import DataFrame

from ..utils import check_directory_exists_and_if_not_mkdir, logger
from .base_sampler import MCMCSampler, SamplerError, signal_wrapper
from .ptemcee import LikePriorEvaluator

_evaluator = LikePriorEvaluator()


class Emcee(MCMCSampler):
    """bilby wrapper emcee (https://github.com/dfm/emcee)

    All positional and keyword arguments (i.e., the args and kwargs) passed to
    `run_sampler` will be propagated to `emcee.EnsembleSampler`, see
    documentation for that class for further help. Under Other Parameters, we
    list commonly used kwargs and the bilby defaults.

    Parameters
    ==========
    nwalkers: int, (500)
        The number of walkers
    nsteps: int, (100)
        The number of steps
    nburn: int (None)
        If given, the fixed number of steps to discard as burn-in. These will
        be discarded from the total number of steps set by `nsteps` and
        therefore the value must be greater than `nsteps`. Else, nburn is
        estimated from the autocorrelation time
    burn_in_fraction: float, (0.25)
        The fraction of steps to discard as burn-in in the event that the
        autocorrelation time cannot be calculated
    burn_in_act: float
        The number of autocorrelation times to discard as burn-in
    a: float (2)
        The proposal scale factor
    verbose: bool
        Whether to print diagnostic information during the analysis

    """

    default_kwargs = dict(
        nwalkers=500,
        a=2,
        args=[],
        kwargs={},
        postargs=None,
        pool=None,
        live_dangerously=False,
        runtime_sortingfn=None,
        lnprob0=None,
        rstate0=None,
        blobs0=None,
        iterations=100,
        thin=1,
        storechain=True,
        mh_proposal=None,
    )

    def __init__(
        self,
        likelihood,
        priors,
        outdir="outdir",
        label="label",
        use_ratio=False,
        plot=False,
        skip_import_verification=False,
        pos0=None,
        nburn=None,
        burn_in_fraction=0.25,
        resume=True,
        burn_in_act=3,
        **kwargs,
    ):
        import emcee

        if LooseVersion(emcee.__version__) > LooseVersion("2.2.1"):
            self.prerelease = True
        else:
            self.prerelease = False
        super(Emcee, self).__init__(
            likelihood=likelihood,
            priors=priors,
            outdir=outdir,
            label=label,
            use_ratio=use_ratio,
            plot=plot,
            skip_import_verification=skip_import_verification,
            **kwargs,
        )
        self._check_version()
        self.resume = resume
        self.pos0 = pos0
        self.nburn = nburn
        self.burn_in_fraction = burn_in_fraction
        self.burn_in_act = burn_in_act
        self.verbose = kwargs.get("verbose", True)

    def _check_version(self):
        import emcee

        if LooseVersion(emcee.__version__) > LooseVersion("2.2.1"):
            self.prerelease = True
        else:
            self.prerelease = False
        return emcee

    def _translate_kwargs(self, kwargs):
        if "nwalkers" not in kwargs:
            for equiv in self.nwalkers_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["nwalkers"] = kwargs.pop(equiv)
        if "iterations" not in kwargs:
            if "nsteps" in kwargs:
                kwargs["iterations"] = kwargs.pop("nsteps")

    @property
    def sampler_function_kwargs(self):
        keys = [
            "lnprob0",
            "rstate0",
            "blobs0",
            "iterations",
            "thin",
            "storechain",
            "mh_proposal",
        ]

        # updated function keywords for emcee > v2.2.1
        updatekeys = {
            "p0": "initial_state",
            "lnprob0": "log_prob0",
            "storechain": "store",
        }

        function_kwargs = {key: self.kwargs[key] for key in keys if key in self.kwargs}
        function_kwargs["p0"] = self.pos0

        if self.prerelease:
            if function_kwargs["mh_proposal"] is not None:
                logger.warning(
                    "The 'mh_proposal' option is no longer used "
                    "in emcee > 2.2.1, and will be ignored."
                )
            del function_kwargs["mh_proposal"]

            for key in updatekeys:
                if updatekeys[key] not in function_kwargs:
                    function_kwargs[updatekeys[key]] = function_kwargs.pop(key)
                else:
                    del function_kwargs[key]

        return function_kwargs

    @property
    def sampler_init_kwargs(self):
        init_kwargs = {
            key: value
            for key, value in self.kwargs.items()
            if key not in self.sampler_function_kwargs
        }

        init_kwargs["lnpostfn"] = _evaluator.call_emcee
        init_kwargs["dim"] = self.ndim

        # updated init keywords for emcee > v2.2.1
        updatekeys = {"dim": "ndim", "lnpostfn": "log_prob_fn"}

        if self.prerelease:
            for key in updatekeys:
                if key in init_kwargs:
                    init_kwargs[updatekeys[key]] = init_kwargs.pop(key)

            oldfunckeys = ["p0", "lnprob0", "storechain", "mh_proposal"]
            for key in oldfunckeys:
                if key in init_kwargs:
                    del init_kwargs[key]

        return init_kwargs

    @property
    def nburn(self):
        if type(self.__nburn) in [float, int]:
            return int(self.__nburn)
        elif self.result.max_autocorrelation_time is None:
            return int(self.burn_in_fraction * self.nsteps)
        else:
            return int(self.burn_in_act * self.result.max_autocorrelation_time)

    @nburn.setter
    def nburn(self, nburn):
        if isinstance(nburn, (float, int)):
            if nburn > self.kwargs["iterations"] - 1:
                raise ValueError(
                    "Number of burn-in samples must be smaller "
                    "than the total number of iterations"
                )

        self.__nburn = nburn

    @property
    def nwalkers(self):
        return self.kwargs["nwalkers"]

    @property
    def nsteps(self):
        return self.kwargs["iterations"]

    @nsteps.setter
    def nsteps(self, nsteps):
        self.kwargs["iterations"] = nsteps

    @property
    def stored_chain(self):
        """Read the stored zero-temperature chain data in from disk"""
        return np.genfromtxt(self.checkpoint_info.chain_file, names=True)

    @property
    def stored_samples(self):
        """Returns the samples stored on disk"""
        return self.stored_chain[self.search_parameter_keys]

    @property
    def stored_loglike(self):
        """Returns the log-likelihood stored on disk"""
        return self.stored_chain["log_l"]

    @property
    def stored_logprior(self):
        """Returns the log-prior stored on disk"""
        return self.stored_chain["log_p"]

    def _init_chain_file(self):
        with open(self.checkpoint_info.chain_file, "w+") as ff:
            search_keys_str = "\t".join(self.search_parameter_keys)
            ff.write(f"walker\t{search_keys_str}\tlog_l\tlog_p\n")

    @property
    def checkpoint_info(self):
        """Defines various things related to checkpointing and storing data

        Returns
        =======
        checkpoint_info: named_tuple
            An object with attributes `sampler_file`, `chain_file`, and
            `chain_template`. The first two give paths to where the sampler and
            chain data is stored, the last a formatted-str-template with which
            to write the chain data to disk

        """
        out_dir = os.path.join(
            self.outdir, f"{self.__class__.__name__.lower()}_{self.label}"
        )
        check_directory_exists_and_if_not_mkdir(out_dir)

        chain_file = os.path.join(out_dir, "chain.dat")
        sampler_file = os.path.join(out_dir, "sampler.pickle")
        chain_template = (
            "{:d}" + "\t{:.9e}" * (len(self.search_parameter_keys) + 2) + "\n"
        )

        CheckpointInfo = namedtuple(
            "CheckpointInfo", ["sampler_file", "chain_file", "chain_template"]
        )

        checkpoint_info = CheckpointInfo(
            sampler_file=sampler_file,
            chain_file=chain_file,
            chain_template=chain_template,
        )

        return checkpoint_info

    @property
    def sampler_chain(self):
        nsteps = self._previous_iterations
        return self.sampler.chain[:, :nsteps, :]

    def write_current_state(self):
        """Writes a pickle file of the sampler to disk using dill"""
        import dill

        logger.info(
            f"Checkpointing sampler to file {self.checkpoint_info.sampler_file}"
        )
        with open(self.checkpoint_info.sampler_file, "wb") as f:
            # Overwrites the stored sampler chain with one that is truncated
            # to only the completed steps
            self.sampler._chain = self.sampler_chain
            _pool = self.sampler.pool
            self.sampler.pool = None
            dill.dump(self._sampler, f)
            self.sampler.pool = _pool

    def _initialise_sampler(self):
        from emcee import EnsembleSampler

        self._sampler = EnsembleSampler(**self.sampler_init_kwargs)
        self._init_chain_file()

    @property
    def sampler(self):
        """Returns the emcee sampler object

        If, already initialized, returns the stored _sampler value. Otherwise,
        first checks if there is a pickle file from which to load. If there is
        not, then initialize the sampler and set the initial random draw

        """
        if hasattr(self, "_sampler"):
            pass
        elif self.resume and os.path.isfile(self.checkpoint_info.sampler_file):
            import dill

            logger.info(
                f"Resuming run from checkpoint file {self.checkpoint_info.sampler_file}"
            )
            with open(self.checkpoint_info.sampler_file, "rb") as f:
                self._sampler = dill.load(f)
                self._sampler.pool = self.pool
            self._set_pos0_for_resume()
        else:
            self._initialise_sampler()
            self._set_pos0()
        return self._sampler

    def write_chains_to_file(self, sample):
        chain_file = self.checkpoint_info.chain_file
        temp_chain_file = chain_file + ".temp"
        if os.path.isfile(chain_file):
            copyfile(chain_file, temp_chain_file)
        if self.prerelease:
            points = np.hstack([sample.coords, sample.blobs])
        else:
            points = np.hstack([sample[0], np.array(sample[3])])
        with open(temp_chain_file, "a") as ff:
            for ii, point in enumerate(points):
                ff.write(self.checkpoint_info.chain_template.format(ii, *point))
        shutil.move(temp_chain_file, chain_file)

    @property
    def _previous_iterations(self):
        """Returns the number of iterations that the sampler has saved

        This is used when loading in a sampler from a pickle file to figure out
        how much of the run has already been completed
        """
        try:
            return len(self.sampler.blobs)
        except AttributeError:
            return 0

    def _draw_pos0_from_prior(self):
        return np.array(
            [self.get_random_draw_from_prior() for _ in range(self.nwalkers)]
        )

    @property
    def _pos0_shape(self):
        return (self.nwalkers, self.ndim)

    def _set_pos0(self):
        if self.pos0 is not None:
            logger.debug("Using given initial positions for walkers")
            if isinstance(self.pos0, DataFrame):
                self.pos0 = self.pos0[self.search_parameter_keys].values
            elif type(self.pos0) in (list, np.ndarray):
                self.pos0 = np.squeeze(self.pos0)

            if self.pos0.shape != self._pos0_shape:
                raise ValueError("Input pos0 should be of shape ndim, nwalkers")
            logger.debug("Checking input pos0")
            for draw in self.pos0:
                self.check_draw(draw)
        else:
            logger.debug("Generating initial walker positions from prior")
            self.pos0 = self._draw_pos0_from_prior()

    def _set_pos0_for_resume(self):
        self.pos0 = self.sampler.chain[:, -1, :]

    @signal_wrapper
    def run_sampler(self):
        self._setup_pool()
        from tqdm.auto import tqdm

        sampler_function_kwargs = self.sampler_function_kwargs
        iterations = sampler_function_kwargs.pop("iterations")
        iterations -= self._previous_iterations

        if self.prerelease:
            sampler_function_kwargs["initial_state"] = self.pos0
        else:
            sampler_function_kwargs["p0"] = self.pos0

        # main iteration loop
        iterator = self.sampler.sample(iterations=iterations, **sampler_function_kwargs)
        if self.verbose:
            iterator = tqdm(iterator, total=iterations)
        for sample in iterator:
            self.write_chains_to_file(sample)
        if self.verbose:
            iterator.close()
        self.write_current_state()

        self.result.sampler_output = np.nan
        self.calculate_autocorrelation(self.sampler.chain.reshape((-1, self.ndim)))
        self.print_nburn_logging_info()

        self._generate_result()

        self.result.samples = self.sampler.chain[:, self.nburn :, :].reshape(
            (-1, self.ndim)
        )
        self.result.walkers = self.sampler.chain
        return self.result

    def _generate_result(self):
        self.result.nburn = self.nburn
        self.calc_likelihood_count()
        if self.result.nburn > self.nsteps:
            raise SamplerError(
                "The run has finished, but the chain is not burned in: "
                f"`nburn < nsteps` ({self.result.nburn} < {self.nsteps})."
                " Try increasing the number of steps."
            )
        blobs = np.array(self.sampler.blobs)
        blobs_trimmed = blobs[self.nburn :, :, :].reshape((-1, 2))
        log_likelihoods, log_priors = blobs_trimmed.T
        self.result.log_likelihood_evaluations = log_likelihoods
        self.result.log_prior_evaluations = log_priors
        self.result.log_evidence = np.nan
        self.result.log_evidence_err = np.nan



import glob
import shutil

import numpy as np

from ..utils import logger
from .base_sampler import MCMCSampler, SamplerNotInstalledError, signal_wrapper


class PTMCMCSampler(MCMCSampler):
    """bilby wrapper of PTMCMC (https://github.com/jellis18/PTMCMCSampler/)

    All positional and keyword arguments (i.e., the args and kwargs) passed to
    `run_sampler` will be propagated to `PTMCMCSampler.PTMCMCSampler`, see
    documentation for that class for further help. Under Other Parameters, we
    list commonly used kwargs and the bilby defaults.

    Parameters
    ==========
    Niter: int (2*10**4 + 1)
        The number of mcmc steps
    burn: int (5 * 10**3)
        If given, the fixed number of steps to discard as burn-in
    thin: int (1)
        The number of steps before saving the sample to the chain
    custom_proposals: dict (None)
        Add dictionary of proposals to the array of proposals, this must be in
        the form of a dictionary with the name of the proposal, then a list
        containing the jump function and the weight e.g {'name' : [function ,
        weight]} see
        (https://github.com/rgreen1995/PTMCMCSampler/blob/master/examples/simple.ipynb)
        and
        (http://jellis18.github.io/PTMCMCSampler/PTMCMCSampler.html#ptmcmcsampler-ptmcmcsampler-module)
        for examples and more info.
    logl_grad: func (None)
        Gradient of likelihood  if known (default = None)
    logp_grad: func (None)
        Gradient of prior if known (default = None)
    verbose: bool (True)
        Update current run-status to the screen

    """

    default_kwargs = {
        "p0": None,
        "Niter": 2 * 10**4 + 1,
        "neff": 10**4,
        "burn": 5 * 10**3,
        "verbose": True,
        "ladder": None,
        "Tmin": 1,
        "Tmax": None,
        "Tskip": 100,
        "isave": 1000,
        "thin": 1,
        "covUpdate": 1000,
        "SCAMweight": 1,
        "AMweight": 1,
        "DEweight": 1,
        "HMCweight": 0,
        "MALAweight": 0,
        "NUTSweight": 0,
        "HMCstepsize": 0.1,
        "HMCsteps": 300,
        "groups": None,
        "custom_proposals": None,
        "loglargs": {},
        "loglkwargs": {},
        "logpargs": {},
        "logpkwargs": {},
        "logl_grad": None,
        "logp_grad": None,
        "outDir": None,
    }
    hard_exit = True

    def __init__(
        self,
        likelihood,
        priors,
        outdir="outdir",
        label="label",
        use_ratio=False,
        plot=False,
        skip_import_verification=False,
        **kwargs,
    ):

        super(PTMCMCSampler, self).__init__(
            likelihood=likelihood,
            priors=priors,
            outdir=outdir,
            label=label,
            use_ratio=use_ratio,
            plot=plot,
            skip_import_verification=skip_import_verification,
            **kwargs,
        )

        if self.kwargs["p0"] is None:
            self.p0 = self.get_random_draw_from_prior()
        else:
            self.p0 = self.kwargs["p0"]
        self.likelihood = likelihood
        self.priors = priors

    def _verify_external_sampler(self):
        # PTMCMC is imported with Caps so need to overwrite the parent function
        # which forces `__name__.lower()
        external_sampler_name = self.__class__.__name__
        try:
            __import__(external_sampler_name)
        except (ImportError, SystemExit):
            raise SamplerNotInstalledError(
                f"Sampler {external_sampler_name} is not installed on this system"
            )

    def _translate_kwargs(self, kwargs):
        if "Niter" not in kwargs:
            for equiv in self.nwalkers_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["Niter"] = kwargs.pop(equiv)
        if "burn" not in kwargs:
            for equiv in self.nburn_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["burn"] = kwargs.pop(equiv)

    @property
    def custom_proposals(self):
        return self.kwargs["custom_proposals"]

    @property
    def sampler_init_kwargs(self):
        keys = [
            "groups",
            "loglargs",
            "logp_grad",
            "logpkwargs",
            "loglkwargs",
            "logl_grad",
            "logpargs",
            "outDir",
            "verbose",
        ]
        init_kwargs = {key: self.kwargs[key] for key in keys}
        if init_kwargs["outDir"] is None:
            init_kwargs["outDir"] = f"{self.outdir}/ptmcmc_temp_{self.label}/"
        return init_kwargs

    @property
    def sampler_function_kwargs(self):
        keys = [
            "Niter",
            "neff",
            "Tmin",
            "HMCweight",
            "covUpdate",
            "SCAMweight",
            "ladder",
            "burn",
            "NUTSweight",
            "AMweight",
            "MALAweight",
            "thin",
            "HMCstepsize",
            "isave",
            "Tskip",
            "HMCsteps",
            "Tmax",
            "DEweight",
        ]
        sampler_kwargs = {key: self.kwargs[key] for key in keys}
        return sampler_kwargs

    @staticmethod
    def _import_external_sampler():
        from PTMCMCSampler import PTMCMCSampler

        return PTMCMCSampler

    @signal_wrapper
    def run_sampler(self):
        PTMCMCSampler = self._import_external_sampler()
        sampler = PTMCMCSampler.PTSampler(
            ndim=self.ndim,
            logp=self.log_prior,
            logl=self.log_likelihood,
            cov=np.eye(self.ndim),
            **self.sampler_init_kwargs,
        )
        if self.custom_proposals is not None:
            for proposal in self.custom_proposals:
                logger.info(
                    f"Adding {proposal} to proposals with weight {self.custom_proposals[proposal][1]}"
                )
                sampler.addProposalToCycle(
                    self.custom_proposals[proposal][0],
                    self.custom_proposals[proposal][1],
                )
        sampler.sample(p0=self.p0, **self.sampler_function_kwargs)
        samples, meta, loglike = self.__read_in_data()

        self.calc_likelihood_count()
        self.result.nburn = self.sampler_function_kwargs["burn"]
        self.result.samples = samples[self.result.nburn :]
        self.meta_data["sampler_meta"] = meta
        self.result.log_likelihood_evaluations = loglike[self.result.nburn :]
        self.result.sampler_output = np.nan
        self.result.walkers = np.nan
        self.result.log_evidence = np.nan
        self.result.log_evidence_err = np.nan
        return self.result

    def __read_in_data(self):
        """Read the data stored by PTMCMC to disk"""
        temp_outDir = self.sampler_init_kwargs["outDir"]
        try:
            data = np.loadtxt(f"{temp_outDir}chain_1.txt")
        except OSError:
            data = np.loadtxt(f"{temp_outDir}chain_1.0.txt")
        jumpfiles = glob.glob(f"{temp_outDir}/*jump.txt")
        jumps = map(np.loadtxt, jumpfiles)
        samples = data[:, :-4]
        loglike = data[:, -3]

        jump_accept = {}
        for ct, j in enumerate(jumps):
            label = jumpfiles[ct].split("/")[-1].split("_jump.txt")[0]
            jump_accept[label] = j
        PT_swap = {"swap_accept": data[:, -1]}
        tot_accept = {"tot_accept": data[:, -2]}
        log_post = {"log_post": data[:, -4]}
        meta = {}
        meta["tot_accept"] = tot_accept
        meta["PT_swap"] = PT_swap
        meta["proposals"] = jump_accept
        meta["log_post"] = log_post

        shutil.rmtree(temp_outDir)

        return samples, meta, loglike

    def write_current_state(self):
        """TODO: implement a checkpointing method"""
        pass



import logging
import os
from pathlib import Path
import sys

logger = logging.getLogger('bilby')


def setup_logger(outdir='.', label=None, log_level='INFO', print_version=False):
    """ Setup logging output: call at the start of the script to use

    Parameters
    ==========
    outdir, label: str
        If supplied, write the logging output to outdir/label.log
    log_level: str, optional
        ['debug', 'info', 'warning']
        Either a string from the list above, or an integer as specified
        in https://docs.python.org/2/library/logging.html#logging-levels
    print_version: bool
        If true, print version information
    """

    if type(log_level) is str:
        try:
            level = getattr(logging, log_level.upper())
        except AttributeError:
            raise ValueError('log_level {} not understood'.format(log_level))
    else:
        level = int(log_level)

    logger = logging.getLogger('bilby')
    logger.propagate = False
    logger.setLevel(level)

    if any([type(h) == logging.StreamHandler for h in logger.handlers]) is False:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter(
            '%(asctime)s %(name)s %(levelname)-8s: %(message)s', datefmt='%H:%M'))
        stream_handler.setLevel(level)
        logger.addHandler(stream_handler)

    if any([type(h) == logging.FileHandler for h in logger.handlers]) is False:
        if label:
            Path(outdir).mkdir(parents=True, exist_ok=True)
            log_file = '{}/{}.log'.format(outdir, label)
            file_handler = logging.FileHandler(log_file)
            file_handler.setFormatter(logging.Formatter(
                '%(asctime)s %(levelname)-8s: %(message)s', datefmt='%H:%M'))

            file_handler.setLevel(level)
            logger.addHandler(file_handler)

    for handler in logger.handlers:
        handler.setLevel(level)

    if print_version:
        version = get_version_information()
        logger.info('Running bilby version: {}'.format(version))


def get_version_information():
    version_file = os.path.join(
        os.path.dirname(os.path.dirname(os.path.dirname(__file__))), '.version')
    try:
        with open(version_file, 'r') as f:
            return f.readline().rstrip()
    except EnvironmentError:
        print("No version information file '.version' found")


def loaded_modules_dict():
    module_names = list(sys.modules.keys())
    vdict = {}
    for key in module_names:
        if "." not in str(key):
            vdict[key] = str(getattr(sys.modules[key], "__version__", "N/A"))
    return vdict



import numpy as np

from ...core.likelihood import Likelihood


class BasicGravitationalWaveTransient(Likelihood):

    def __init__(self, interferometers, waveform_generator):
        """

        A likelihood object, able to compute the likelihood of the data given
        some model parameters

        The simplest frequency-domain gravitational wave transient likelihood. Does
        not include distance/phase marginalization.


        Parameters
        ==========
        interferometers: list
            A list of `bilby.gw.detector.Interferometer` instances - contains the
            detector data and power spectral densities
        waveform_generator: bilby.gw.waveform_generator.WaveformGenerator
            An object which computes the frequency-domain strain of the signal,
            given some set of parameters

        """
        super(BasicGravitationalWaveTransient, self).__init__(dict())
        self.interferometers = interferometers
        self.waveform_generator = waveform_generator

    def __repr__(self):
        return self.__class__.__name__ + '(interferometers={},\n\twaveform_generator={})' \
            .format(self.interferometers, self.waveform_generator)

    def noise_log_likelihood(self):
        """ Calculates the real part of noise log-likelihood

        Returns
        =======
        float: The real part of the noise log likelihood

        """
        log_l = 0
        for interferometer in self.interferometers:
            log_l -= 2. / self.waveform_generator.duration * np.sum(
                abs(interferometer.frequency_domain_strain) ** 2 /
                interferometer.power_spectral_density_array)
        return log_l.real

    def log_likelihood(self):
        """ Calculates the real part of log-likelihood value

        Returns
        =======
        float: The real part of the log likelihood

        """
        log_l = 0
        waveform_polarizations = \
            self.waveform_generator.frequency_domain_strain(
                self.parameters.copy())
        if waveform_polarizations is None:
            return np.nan_to_num(-np.inf)
        for interferometer in self.interferometers:
            log_l += self.log_likelihood_interferometer(
                waveform_polarizations, interferometer)
        return log_l.real

    def log_likelihood_interferometer(self, waveform_polarizations,
                                      interferometer):
        """

        Parameters
        ==========
        waveform_polarizations: dict
            Dictionary containing the desired waveform polarization modes and the related strain
        interferometer: bilby.gw.detector.Interferometer
            The Interferometer object we want to have the log-likelihood for

        Returns
        =======
        float: The real part of the log-likelihood for this interferometer

        """
        signal_ifo = interferometer.get_detector_response(
            waveform_polarizations, self.parameters)

        log_l = - 2. / self.waveform_generator.duration * np.vdot(
            interferometer.frequency_domain_strain - signal_ifo,
            (interferometer.frequency_domain_strain - signal_ifo) /
            interferometer.power_spectral_density_array)
        return log_l.real



import multiprocessing
import os
import sys
import threading
import time
from signal import SIGINT

multiprocessing.set_start_method("fork")  # noqa

import unittest
import pytest
from parameterized import parameterized
import shutil

import bilby
import numpy as np


_sampler_kwargs = dict(
    bilby_mcmc=dict(nsamples=200, printdt=1),
    cpnest=dict(nlive=100),
    dnest4=dict(
        max_num_levels=2,
        num_steps=10,
        new_level_interval=10,
        num_per_step=10,
        thread_steps=1,
        num_particles=50,
        max_pool=1,
    ),
    dynesty=dict(nlive=100),
    dynamic_dynesty=dict(
        nlive_init=100,
        nlive_batch=100,
        dlogz_init=1.0,
        maxbatch=0,
        maxcall=100,
        bound="single",
    ),
    emcee=dict(iterations=1000, nwalkers=10),
    kombine=dict(iterations=200, nwalkers=10, autoburnin=False),
    nessai=dict(
        nlive=100,
        poolsize=1000,
        max_iteration=1000,
        max_threads=3,
    ),
    nestle=dict(nlive=100),
    ptemcee=dict(
        nsamples=100,
        nwalkers=50,
        burn_in_act=1,
        ntemps=1,
        frac_threshold=0.5,
    ),
    PTMCMCSampler=dict(Niter=101, burn=2, isave=100),
    # pymc3=dict(draws=50, tune=50, n_init=250),  removed until testing issue can be resolved
    pymultinest=dict(nlive=100),
    pypolychord=dict(nlive=100),
    ultranest=dict(nlive=100, temporary_directory=False),
)

sampler_imports = dict(
    bilby_mcmc="bilby",
    dynamic_dynesty="dynesty"
)

no_pool_test = ["dnest4", "pymultinest", "nestle", "ptmcmcsampler", "pypolychord", "ultranest"]


def slow_func(x, m, c):
    time.sleep(0.01)
    return m * x + c


def model(x, m, c):
    return m * x + c


class TestRunningSamplers(unittest.TestCase):
    def setUp(self):
        np.random.seed(42)
        bilby.core.utils.command_line_args.bilby_test_mode = False
        self.x = np.linspace(0, 1, 11)
        self.injection_parameters = dict(m=0.5, c=0.2)
        self.sigma = 0.1
        self.y = model(self.x, **self.injection_parameters) + np.random.normal(
            0, self.sigma, len(self.x)
        )
        self.likelihood = bilby.likelihood.GaussianLikelihood(
            self.x, self.y, model, self.sigma
        )

        self.priors = bilby.core.prior.PriorDict()
        self.priors["m"] = bilby.core.prior.Uniform(0, 5, boundary="periodic")
        self.priors["c"] = bilby.core.prior.Uniform(-2, 2, boundary="reflective")
        self._remove_tree()
        bilby.core.utils.check_directory_exists_and_if_not_mkdir("outdir")

    @staticmethod
    def conversion_function(parameters, likelihood, prior):
        converted = parameters.copy()
        if "derived" not in converted:
            converted["derived"] = converted["m"] * converted["c"]
        return converted

    def tearDown(self):
        del self.likelihood
        del self.priors
        bilby.core.utils.command_line_args.bilby_test_mode = False
        self._remove_tree()

    def _remove_tree(self):
        try:
            shutil.rmtree("outdir")
        except OSError:
            pass

    @parameterized.expand(_sampler_kwargs.keys())
    def test_run_sampler_single(self, sampler):
        self._run_sampler(sampler, pool_size=1)

    @parameterized.expand(_sampler_kwargs.keys())
    def test_run_sampler_pool(self, sampler):
        self._run_sampler(sampler, pool_size=2)

    def _run_sampler(self, sampler, pool_size, **extra_kwargs):
        pytest.importorskip(sampler_imports.get(sampler, sampler))
        if pool_size > 1 and sampler.lower() in no_pool_test:
            pytest.skip(f"{sampler} cannot be parallelized")
        bilby.core.utils.check_directory_exists_and_if_not_mkdir("outdir")
        kwargs = _sampler_kwargs[sampler]
        res = bilby.run_sampler(
            likelihood=self.likelihood,
            priors=self.priors,
            sampler=sampler,
            save=False,
            npool=pool_size,
            conversion_function=self.conversion_function,
            **kwargs,
            **extra_kwargs,
        )
        assert "derived" in res.posterior
        assert res.log_likelihood_evaluations is not None

    @parameterized.expand(_sampler_kwargs.keys())
    def test_interrupt_sampler_single(self, sampler):
        self._run_with_signal_handling(sampler, pool_size=1)

    @parameterized.expand(_sampler_kwargs.keys())
    def test_interrupt_sampler_pool(self, sampler):
        self._run_with_signal_handling(sampler, pool_size=2)

    def _run_with_signal_handling(self, sampler, pool_size=1):
        pytest.importorskip(sampler_imports.get(sampler, sampler))
        if bilby.core.sampler.IMPLEMENTED_SAMPLERS[sampler.lower()].hard_exit:
            pytest.skip(f"{sampler} hard exits, can't test signal handling.")
        if pool_size > 1 and sampler.lower() in no_pool_test:
            pytest.skip(f"{sampler} cannot be parallelized")
        if sys.version_info.minor == 8 and sampler.lower == "cpnest":
            pytest.skip("Pool interrupting broken for cpnest with py3.8")
        if sampler.lower() == "nessai" and pool_size > 1:
            pytest.skip(
                "Interrupting with a pool is failing in pytest. "
                "Likely due to interactions with the signal handling in nessai."
            )
        pid = os.getpid()
        print(sampler)

        def trigger_signal():
            # You could do something more robust, e.g. wait until port is listening
            time.sleep(4)
            os.kill(pid, SIGINT)

        thread = threading.Thread(target=trigger_signal)
        thread.daemon = True
        thread.start()

        self.likelihood._func = slow_func

        with self.assertRaises((SystemExit, KeyboardInterrupt)):
            try:
                while True:
                    self._run_sampler(sampler=sampler, pool_size=pool_size, exit_code=5)
            except SystemExit as error:
                self.assertEqual(error.code, 5)
                raise


if __name__ == "__main__":
    unittest.main()




import json

import numpy as np

from .base import GravitationalWaveTransient
from ...core.utils import BilbyJsonEncoder, decode_bilby_json
from ...core.utils import (
    logger, create_frequency_series, speed_of_light, radius_of_earth
)
from ..prior import CBCPriorDict


class ROQGravitationalWaveTransient(GravitationalWaveTransient):
    """A reduced order quadrature likelihood object

    This uses the method described in Smith et al., (2016) Phys. Rev. D 94,
    044031. A public repository of the ROQ data is available from
    https://git.ligo.org/lscsoft/ROQ_data.

    Parameters
    ==========
    interferometers: list, bilby.gw.detector.InterferometerList
        A list of `bilby.detector.Interferometer` instances - contains the
        detector data and power spectral densities
    waveform_generator: `bilby.waveform_generator.WaveformGenerator`
        An object which computes the frequency-domain strain of the signal,
        given some set of parameters
    linear_matrix: str, array_like
        Either a string point to the file from which to load the linear_matrix
        array, or the array itself.
    quadratic_matrix: str, array_like
        Either a string point to the file from which to load the
        quadratic_matrix array, or the array itself.
    roq_params: str, array_like
        Parameters describing the domain of validity of the ROQ basis.
    roq_params_check: bool
        If true, run tests using the roq_params to check the prior and data are
        valid for the ROQ
    roq_scale_factor: float
        The ROQ scale factor used.
    parameter_conversion: func, optional
        Function to update self.parameters before bases are selected based on
        the values of self.parameters. This enables a user to switch bases
        based on the values of parameters which are not directly used for
        sampling.
    priors: dict, bilby.prior.PriorDict
        A dictionary of priors containing at least the geocent_time prior
        Warning: when using marginalisation the dict is overwritten which will change the
        the dict you are passing in. If this behaviour is undesired, pass `priors.copy()`.
    time_marginalization: bool, optional
        If true, marginalize over time in the likelihood.
        The spacing of time samples can be specified through delta_tc.
        If using time marginalisation and jitter_time is True a "jitter"
        parameter is added to the prior which modifies the position of the
        grid of times.
    jitter_time: bool, optional
        Whether to introduce a `time_jitter` parameter. This avoids either
        missing the likelihood peak, or introducing biases in the
        reconstructed time posterior due to an insufficient sampling frequency.
        Default is False, however using this parameter is strongly encouraged.
    delta_tc: float, optional
        The spacing of time samples for time marginalization. If not specified,
        it is determined based on the signal-to-noise ratio of signal.
    distance_marginalization_lookup_table: (dict, str), optional
        If a dict, dictionary containing the lookup_table, distance_array,
        (distance) prior_array, and reference_distance used to construct
        the table.
        If a string the name of a file containing these quantities.
        The lookup table is stored after construction in either the
        provided string or a default location:
        '.distance_marginalization_lookup_dmin{}_dmax{}_n{}.npz'
    reference_frame: (str, bilby.gw.detector.InterferometerList, list), optional
        Definition of the reference frame for the sky location.
        - "sky": sample in RA/dec, this is the default
        - e.g., "H1L1", ["H1", "L1"], InterferometerList(["H1", "L1"]):
          sample in azimuth and zenith, `azimuth` and `zenith` defined in the
          frame where the z-axis is aligned the the vector connecting H1
          and L1.
    time_reference: str, optional
        Name of the reference for the sampled time parameter.
        - "geocent"/"geocenter": sample in the time at the Earth's center,
          this is the default
        - e.g., "H1": sample in the time of arrival at H1

    """
    def __init__(
            self, interferometers, waveform_generator, priors,
            weights=None, linear_matrix=None, quadratic_matrix=None,
            roq_params=None, roq_params_check=True, roq_scale_factor=1,
            distance_marginalization=False, phase_marginalization=False,
            time_marginalization=False, jitter_time=True, delta_tc=None,
            distance_marginalization_lookup_table=None,
            reference_frame="sky", time_reference="geocenter",
            parameter_conversion=None

    ):
        self._delta_tc = delta_tc
        super(ROQGravitationalWaveTransient, self).__init__(
            interferometers=interferometers,
            waveform_generator=waveform_generator, priors=priors,
            distance_marginalization=distance_marginalization,
            phase_marginalization=phase_marginalization,
            time_marginalization=time_marginalization,
            distance_marginalization_lookup_table=distance_marginalization_lookup_table,
            jitter_time=jitter_time,
            reference_frame=reference_frame,
            time_reference=time_reference
        )

        self.roq_params_check = roq_params_check
        self.roq_scale_factor = roq_scale_factor
        if isinstance(roq_params, np.ndarray) or roq_params is None:
            self.roq_params = roq_params
        elif isinstance(roq_params, str):
            self.roq_params_file = roq_params
            self.roq_params = np.genfromtxt(roq_params, names=True)
        else:
            raise TypeError("roq_params should be array or str")
        if isinstance(weights, dict):
            self.weights = weights
        elif isinstance(weights, str):
            self.weights = self.load_weights(weights)
        else:
            is_hdf5_linear = isinstance(linear_matrix, str) and linear_matrix.endswith('.hdf5')
            linear_matrix = self._parse_basis(linear_matrix, 'linear')
            is_hdf5_quadratic = isinstance(quadratic_matrix, str) and quadratic_matrix.endswith('.hdf5')
            quadratic_matrix = self._parse_basis(quadratic_matrix, 'quadratic')
            # retrieve roq params from a basis file if it is .hdf5
            if self.roq_params is None:
                if is_hdf5_linear:
                    self.roq_params = np.array(
                        [(linear_matrix['minimum_frequency_hz'][()],
                          linear_matrix['maximum_frequency_hz'][()],
                          linear_matrix['duration_s'][()])],
                        dtype=[('flow', float), ('fhigh', float), ('seglen', float)]
                    )
                elif is_hdf5_quadratic:
                    self.roq_params = np.array(
                        [(quadratic_matrix['minimum_frequency_hz'][()],
                          quadratic_matrix['maximum_frequency_hz'][()],
                          quadratic_matrix['duration_s'][()])],
                        dtype=[('flow', float), ('fhigh', float), ('seglen', float)]
                    )
            self.weights = dict()
            self._set_weights(linear_matrix=linear_matrix, quadratic_matrix=quadratic_matrix)
            if is_hdf5_linear:
                linear_matrix.close()
            if is_hdf5_quadratic:
                quadratic_matrix.close()

        self.number_of_bases_linear = len(self.weights[f'{self.interferometers[0].name}_linear'])
        self.number_of_bases_quadratic = len(self.weights[f'{self.interferometers[0].name}_quadratic'])
        self._cache = dict(parameters=None, basis_number_linear=None, basis_number_quadratic=None)
        self.parameter_conversion = parameter_conversion

        for basis_type in ['linear', 'quadratic']:
            number_of_bases = getattr(self, f'number_of_bases_{basis_type}')
            if number_of_bases > 1:
                self._verify_prior_ranges_and_frequency_nodes(basis_type)
            else:
                self._check_frequency_nodes_exist(basis_type)

    def _verify_prior_ranges_and_frequency_nodes(self, basis_type):
        """
        Check if self.weights contains lists of prior ranges and frequency nodes, and their sizes are equal to the
        number of bases.

        Parameters
        ==========
        basis_type: str

        """
        number_of_bases = getattr(self, f'number_of_bases_{basis_type}')
        key = f'prior_range_{basis_type}'
        try:
            prior_ranges = self.weights[key]
        except KeyError:
            raise AttributeError(
                f'For the use of multiple {basis_type} ROQ bases, weights should contain "{key}".')
        else:
            for param_name in prior_ranges:
                if len(prior_ranges[param_name]) != number_of_bases:
                    raise ValueError(
                        f'The number of prior ranges for "{param_name}" does not '
                        f'match the number of {basis_type} bases')
        key = f'frequency_nodes_{basis_type}'
        try:
            frequency_nodes = self.weights[key]
        except KeyError:
            raise AttributeError(
                f'For the use of multiple {basis_type} ROQ bases, weights should contain "{key}".')
        else:
            if len(frequency_nodes) != number_of_bases:
                raise ValueError(
                    f'The number of arrays of frequency nodes does not match the number of {basis_type} bases')

    def _check_frequency_nodes_exist(self, basis_type):
        """
        Check if self._waveform_generator.waveform_arguments contains frequency nodes. If not, they are retrieved from
        self.weights. If self.weights neither, raise AttributeError.

        Parameters
        ==========
        basis_type: str

        """
        key = f'frequency_nodes_{basis_type}'
        if key not in self._waveform_generator.waveform_arguments:
            try:
                self._waveform_generator.waveform_arguments[key] = self.weights[key][0]
            except KeyError:
                raise AttributeError(
                    f'{key} should be contained in weights or waveform arguments.')

    def _setup_time_marginalization(self):
        if self._delta_tc is None:
            self._delta_tc = self._get_time_resolution()
        tcmin = self.priors['geocent_time'].minimum
        tcmax = self.priors['geocent_time'].maximum
        number_of_time_samples = int(np.ceil((tcmax - tcmin) / self._delta_tc))
        # adjust delta tc so that the last time sample has an equal weight
        self._delta_tc = (tcmax - tcmin) / number_of_time_samples
        logger.info(
            "delta tc for time marginalization = {} seconds.".format(self._delta_tc))
        self._times = tcmin + self._delta_tc / 2. + np.arange(number_of_time_samples) * self._delta_tc
        self._beam_pattern_reference_time = (tcmin + tcmax) / 2.

    @staticmethod
    def _parse_basis(basis, basis_type):
        """
        Parse basis and format it to an hdf5-like object

        Parameters
        ----------
        basis : array-like or str
            array-like basis or path to file
        basis_type : str
            'linear' or 'quadratic'

        Returns
        -------
        basis : hdf5-like object

        """
        if basis_type not in ['linear', 'quadratic']:
            raise ValueError(f'basis_type {basis_type} not recognized')
        if isinstance(basis, str):
            logger.info(f'Loading {basis_type}_matrix from {basis}')
            format = basis.split('.')[-1]
            if format == 'npy':
                basis = {f'basis_{basis_type}': {'0': {'basis': np.load(basis)}}}
            elif format == 'hdf5':
                import h5py
                basis = h5py.File(basis, 'r')
            else:
                raise IOError(f'Format {format} not recognized.')
        elif isinstance(basis, np.ndarray):
            basis = {f'basis_{basis_type}': {'0': {'basis': basis.T}}}
        else:
            raise TypeError('basis needs to be str or np.ndarray')
        return basis

    def _select_prior_ranges(self, prior_ranges):
        """
        Select prior ranges which have intersection with self.priors

        Parameters
        ----------
        prior_ranges : dict
            dictionary whose keys are parameter names and values are ndarray of
            their prior ranges

        Returns
        -------
        idxs_in_prior_range : ndarray
            indexes of selected prior ranges
        selected_prior_ranges : dict

        """
        param_names = list(prior_ranges.keys())
        number_of_prior_ranges = len(prior_ranges[param_names[0]])
        in_prior_range = np.ones(number_of_prior_ranges, dtype=bool)
        for param_name in param_names:
            try:
                prior = self.priors[param_name]
            except KeyError:
                continue
            prior_ranges_of_this_param = prior_ranges[param_name]
            in_prior_range *= \
                (prior_ranges_of_this_param[:, 1] >= prior.minimum) * \
                (prior_ranges_of_this_param[:, 0] <= prior.maximum)
        idxs_in_prior_range = np.arange(number_of_prior_ranges)[in_prior_range]
        return idxs_in_prior_range, \
            dict((param_name, prior_ranges[param_name][idxs_in_prior_range])
                 for param_name in param_names)

    def _update_basis(self):
        """
        Update basis and frequency nodes depending on the curret values of parameters

        This updates
        - self._cache
        - frequency_nodes_linear/quadratic in self._waveform_generator.waveform_arguments

        """
        parameters = self.parameters.copy()
        if self.parameter_conversion is not None:
            parameters = self.parameter_conversion(parameters)
        for basis_type, number_of_bases in zip(
            ['linear', 'quadratic'], [self.number_of_bases_linear, self.number_of_bases_quadratic]
        ):
            if number_of_bases == 1:
                continue
            in_prior_range = np.ones(number_of_bases, dtype=bool)
            prior_range_key = f'prior_range_{basis_type}'
            for param_name in self.weights[prior_range_key]:
                if param_name not in parameters:
                    continue
                in_prior_range *= \
                    (self.weights[prior_range_key][param_name][:, 0] <= parameters[param_name]) * \
                    (self.weights[prior_range_key][param_name][:, 1] >= parameters[param_name])
            basis_number_key = f'basis_number_{basis_type}'
            self._cache[basis_number_key] = np.arange(number_of_bases)[in_prior_range][0]
            frequency_nodes_key = f'frequency_nodes_{basis_type}'
            self._waveform_generator.waveform_arguments[frequency_nodes_key] = \
                self.weights[frequency_nodes_key][self._cache[basis_number_key]]
        self._cache['parameters'] = self.parameters.copy()

    @property
    def basis_number_linear(self):
        if self.number_of_bases_linear > 1:
            if self.parameters != self._cache['parameters']:
                self._update_basis()
            return self._cache['basis_number_linear']
        else:
            return 0

    @property
    def basis_number_quadratic(self):
        if self.number_of_bases_quadratic > 1:
            if self.parameters != self._cache['parameters']:
                self._update_basis()
            return self._cache['basis_number_quadratic']
        else:
            return 0

    @property
    def waveform_generator(self):
        if getattr(self, 'number_of_bases_linear', 1) > 1 or getattr(self, 'number_of_bases_quadratic', 1) > 1:
            if self.parameters != self._cache['parameters']:
                self._update_basis()
        return self._waveform_generator

    @waveform_generator.setter
    def waveform_generator(self, waveform_generator):
        self._waveform_generator = waveform_generator

    def calculate_snrs(self, waveform_polarizations, interferometer):
        """
        Compute the snrs for ROQ

        Parameters
        ==========
        waveform_polarizations: waveform
        interferometer: bilby.gw.detector.Interferometer

        """
        if self.time_marginalization:
            time_ref = self._beam_pattern_reference_time
        else:
            time_ref = self.parameters['geocent_time']

        frequency_nodes_linear = self.waveform_generator.waveform_arguments['frequency_nodes_linear']
        frequency_nodes_quadratic = self.waveform_generator.waveform_arguments['frequency_nodes_quadratic']
        size_linear = len(frequency_nodes_linear)
        size_quadratic = len(frequency_nodes_quadratic)
        h_linear = np.zeros(size_linear, dtype=complex)
        h_quadratic = np.zeros(size_quadratic, dtype=complex)
        for mode in waveform_polarizations['linear']:
            response = interferometer.antenna_response(
                self.parameters['ra'], self.parameters['dec'],
                self.parameters['geocent_time'], self.parameters['psi'],
                mode
            )
            h_linear += waveform_polarizations['linear'][mode] * response
            h_quadratic += waveform_polarizations['quadratic'][mode] * response

        calib_linear = interferometer.calibration_model.get_calibration_factor(
            frequency_nodes_linear, prefix='recalib_{}_'.format(interferometer.name), **self.parameters)
        calib_quadratic = interferometer.calibration_model.get_calibration_factor(
            frequency_nodes_quadratic, prefix='recalib_{}_'.format(interferometer.name), **self.parameters)

        h_linear *= calib_linear
        h_quadratic *= calib_quadratic

        optimal_snr_squared = np.vdot(
            np.abs(h_quadratic)**2,
            self.weights[interferometer.name + '_quadratic'][self.basis_number_quadratic]
        )

        dt = interferometer.time_delay_from_geocenter(
            self.parameters['ra'], self.parameters['dec'], time_ref)

        if not self.time_marginalization:
            dt_geocent = self.parameters['geocent_time'] - interferometer.strain_data.start_time
            ifo_time = dt_geocent + dt

            indices, in_bounds = self._closest_time_indices(
                ifo_time, self.weights['time_samples'])
            if not in_bounds:
                logger.debug("SNR calculation error: requested time at edge of ROQ time samples")
                return self._CalculatedSNRs(
                    d_inner_h=np.nan_to_num(-np.inf), optimal_snr_squared=0,
                    complex_matched_filter_snr=np.nan_to_num(-np.inf),
                    d_inner_h_squared_tc_array=None,
                    d_inner_h_array=None,
                    optimal_snr_squared_array=None)

            d_inner_h_tc_array = np.einsum(
                'i,ji->j', np.conjugate(h_linear),
                self.weights[interferometer.name + '_linear'][self.basis_number_linear][indices])

            d_inner_h = self._interp_five_samples(
                self.weights['time_samples'][indices], d_inner_h_tc_array, ifo_time)

            with np.errstate(invalid="ignore"):
                complex_matched_filter_snr = d_inner_h / (optimal_snr_squared**0.5)

            d_inner_h_array = None

        else:
            ifo_times = self._times - interferometer.strain_data.start_time + dt
            if self.jitter_time:
                ifo_times += self.parameters['time_jitter']
            d_inner_h_array = self._calculate_d_inner_h_array(ifo_times, h_linear, interferometer.name)

            d_inner_h = 0.
            complex_matched_filter_snr = 0.

        return self._CalculatedSNRs(
            d_inner_h=d_inner_h, optimal_snr_squared=optimal_snr_squared,
            complex_matched_filter_snr=complex_matched_filter_snr,
            d_inner_h_squared_tc_array=None,
            d_inner_h_array=d_inner_h_array,
            optimal_snr_squared_array=None)

    @staticmethod
    def _closest_time_indices(time, samples):
        """
        Get the closest five times

        Parameters
        ==========
        time: float
            Time to check
        samples: array-like
            Available times

        Returns
        =======
        indices: list
            Indices nearest to time
        in_bounds: bool
            Whether the indices are for valid times
        """
        closest = int((time - samples[0]) / (samples[1] - samples[0]))
        indices = [closest + ii for ii in [-2, -1, 0, 1, 2]]
        in_bounds = (indices[0] >= 0) & (indices[-1] < samples.size)
        return indices, in_bounds

    @staticmethod
    def _interp_five_samples(time_samples, values, time):
        """
        Interpolate a function of time with its values at the closest five times.
        The algorithm is explained in https://dcc.ligo.org/T2100224.

        Parameters
        ==========
        time_samples: array-like
            Closest 5 times
        values: array-like
            The values of the function at closest 5 times
        time: float
            Time at which the function is calculated

        Returns
        =======
        value: float
            The value of the function at the input time
        """
        r1 = (-values[0] + 8. * values[1] - 14. * values[2] + 8. * values[3] - values[4]) / 4.
        r2 = values[2] - 2. * values[3] + values[4]
        a = (time_samples[3] - time) / (time_samples[1] - time_samples[0])
        b = 1. - a
        c = (a**3. - a) / 6.
        d = (b**3. - b) / 6.
        return a * values[2] + b * values[3] + c * r1 + d * r2

    def _calculate_d_inner_h_array(self, times, h_linear, ifo_name):
        """
        Calculate d_inner_h at regularly-spaced time samples. Each value is
        interpolated from the nearest 5 samples with the algorithm explained in
        https://dcc.ligo.org/T2100224.

        Parameters
        ==========
        times: array-like
            Regularly-spaced time samples at which d_inner_h are calculated.
        h_linear: array-like
            Waveforms at linear frequency nodes
        ifo_name: str

        Returns
        =======
        d_inner_h_array: array-like
        """
        roq_time_space = self.weights['time_samples'][1] - self.weights['time_samples'][0]
        times_per_roq_time_space = (times - self.weights['time_samples'][0]) / roq_time_space
        closest_idxs = np.floor(times_per_roq_time_space).astype(int)
        # Get the nearest 5 samples of d_inner_h. Calculate only the required d_inner_h values if the time
        # spacing is larger than 5 times the ROQ time spacing.
        weights_linear = self.weights[ifo_name + '_linear'][self.basis_number_linear]
        h_linear_conj = np.conjugate(h_linear)
        if (times[1] - times[0]) / roq_time_space > 5:
            d_inner_h_m2 = np.dot(weights_linear[closest_idxs - 2], h_linear_conj)
            d_inner_h_m1 = np.dot(weights_linear[closest_idxs - 1], h_linear_conj)
            d_inner_h_0 = np.dot(weights_linear[closest_idxs], h_linear_conj)
            d_inner_h_p1 = np.dot(weights_linear[closest_idxs + 1], h_linear_conj)
            d_inner_h_p2 = np.dot(weights_linear[closest_idxs + 2], h_linear_conj)
        else:
            d_inner_h_at_roq_time_samples = np.dot(weights_linear, h_linear_conj)
            d_inner_h_m2 = d_inner_h_at_roq_time_samples[closest_idxs - 2]
            d_inner_h_m1 = d_inner_h_at_roq_time_samples[closest_idxs - 1]
            d_inner_h_0 = d_inner_h_at_roq_time_samples[closest_idxs]
            d_inner_h_p1 = d_inner_h_at_roq_time_samples[closest_idxs + 1]
            d_inner_h_p2 = d_inner_h_at_roq_time_samples[closest_idxs + 2]
        # quantities required for spline interpolation
        b = times_per_roq_time_space - closest_idxs
        a = 1. - b
        c = (a**3. - a) / 6.
        d = (b**3. - b) / 6.
        r1 = (-d_inner_h_m2 + 8. * d_inner_h_m1 - 14. * d_inner_h_0 + 8. * d_inner_h_p1 - d_inner_h_p2) / 4.
        r2 = d_inner_h_0 - 2. * d_inner_h_p1 + d_inner_h_p2
        return a * d_inner_h_0 + b * d_inner_h_p1 + c * r1 + d * r2

    def perform_roq_params_check(self, ifo=None):
        """ Perform checking that the prior and data are valid for the ROQ

        Parameters
        ==========
        ifo: bilby.gw.detector.Interferometer
            The interferometer
        """
        if self.roq_params_check is False:
            logger.warning("No ROQ params checking performed")
            return
        else:
            if getattr(self, "roq_params_file", None) is not None:
                msg = ("Check ROQ params {} with roq_scale_factor={}"
                       .format(self.roq_params_file, self.roq_scale_factor))
            else:
                msg = ("Check ROQ params with roq_scale_factor={}"
                       .format(self.roq_scale_factor))
            logger.info(msg)

        roq_params = self.roq_params
        roq_minimum_frequency = roq_params['flow'] * self.roq_scale_factor
        roq_maximum_frequency = roq_params['fhigh'] * self.roq_scale_factor
        roq_segment_length = roq_params['seglen'] / self.roq_scale_factor
        try:
            roq_minimum_chirp_mass = roq_params['chirpmassmin'] / self.roq_scale_factor
        except ValueError:
            roq_minimum_chirp_mass = None
        try:
            roq_maximum_chirp_mass = roq_params['chirpmassmax'] / self.roq_scale_factor
        except ValueError:
            roq_maximum_chirp_mass = None
        try:
            roq_minimum_component_mass = roq_params['compmin'] / self.roq_scale_factor
        except ValueError:
            roq_minimum_component_mass = None

        if ifo.maximum_frequency > roq_maximum_frequency:
            raise BilbyROQParamsRangeError(
                "Requested maximum frequency {} larger than ROQ basis fhigh {}"
                .format(ifo.maximum_frequency, roq_maximum_frequency)
            )
        if ifo.minimum_frequency < roq_minimum_frequency:
            raise BilbyROQParamsRangeError(
                "Requested minimum frequency {} lower than ROQ basis flow {}"
                .format(ifo.minimum_frequency, roq_minimum_frequency)
            )
        if ifo.strain_data.duration != roq_segment_length:
            raise BilbyROQParamsRangeError(
                "Requested duration differs from ROQ basis seglen")

        priors = self.priors
        if isinstance(priors, CBCPriorDict) is False:
            logger.warning("Unable to check ROQ parameter bounds: priors not understood")
            return

        if roq_minimum_chirp_mass is not None:
            if priors.minimum_chirp_mass is None:
                logger.warning("Unable to check minimum chirp mass ROQ bounds")
            elif priors.minimum_chirp_mass < roq_minimum_chirp_mass:
                raise BilbyROQParamsRangeError(
                    "Prior minimum chirp mass {} less than ROQ basis bound {}"
                    .format(priors.minimum_chirp_mass, roq_minimum_chirp_mass)
                )

        if roq_maximum_chirp_mass is not None:
            if priors.maximum_chirp_mass is None:
                logger.warning("Unable to check maximum_chirp mass ROQ bounds")
            elif priors.maximum_chirp_mass > roq_maximum_chirp_mass:
                raise BilbyROQParamsRangeError(
                    "Prior maximum chirp mass {} greater than ROQ basis bound {}"
                    .format(priors.maximum_chirp_mass, roq_maximum_chirp_mass)
                )

        if roq_minimum_component_mass is not None:
            if priors.minimum_component_mass is None:
                logger.warning("Unable to check minimum component mass ROQ bounds")
            elif priors.minimum_component_mass < roq_minimum_component_mass:
                raise BilbyROQParamsRangeError(
                    "Prior minimum component mass {} less than ROQ basis bound {}"
                    .format(priors.minimum_component_mass, roq_minimum_component_mass)
                )

    def _set_weights(self, linear_matrix, quadratic_matrix):
        """
        Setup the time-dependent ROQ weights.

        Parameters
        ==========
        linear_matrix, quadratic_matrix: dictionary or h5py.File
            linear and quadratic basis

        """
        time_space = self._get_time_resolution()
        number_of_time_samples = int(self.interferometers.duration / time_space)
        earth_light_crossing_time = 2 * radius_of_earth / speed_of_light + 5 * time_space
        start_idx = max(
            0,
            int(np.floor((
                self.priors['{}_time'.format(self.time_reference)].minimum
                - earth_light_crossing_time
                - self.interferometers.start_time
            ) / time_space))
        )
        end_idx = min(
            number_of_time_samples - 1,
            int(np.ceil((
                self.priors['{}_time'.format(self.time_reference)].maximum
                + earth_light_crossing_time
                - self.interferometers.start_time
            ) / time_space))
        )
        self.weights['time_samples'] = np.arange(start_idx, end_idx + 1) * time_space
        logger.info("Using {} ROQ time samples".format(len(self.weights['time_samples'])))

        # select bases to be used, set prior ranges and frequency nodes if exist
        idxs_in_prior_range = dict()
        for basis_type, matrix in zip(['linear', 'quadratic'], [linear_matrix, quadratic_matrix]):
            key = f'prior_range_{basis_type}'
            if key in matrix:
                prior_ranges = {}
                for param_name in matrix[key]:
                    if 'roq_scale_power' in matrix[key][param_name].attrs:
                        roq_scale_factor = self.roq_scale_factor**matrix[key][param_name].attrs['roq_scale_power']
                    else:
                        roq_scale_factor = 1.
                    prior_ranges[param_name] = matrix[key][param_name][()] * roq_scale_factor
                selected_idxs, selected_prior_ranges = self._select_prior_ranges(prior_ranges)
                self.weights[key] = selected_prior_ranges
                idxs_in_prior_range[basis_type] = selected_idxs
            else:
                idxs_in_prior_range[basis_type] = [0]
            if 'frequency_nodes' in matrix[f'basis_{basis_type}'][str(idxs_in_prior_range[basis_type][0])]:
                self.weights[f'frequency_nodes_{basis_type}'] = [
                    matrix[f'basis_{basis_type}'][str(i)]['frequency_nodes'][()] * self.roq_scale_factor
                    for i in idxs_in_prior_range[basis_type]]

        if 'multiband_linear' in linear_matrix:
            multiband_linear = linear_matrix['multiband_linear'][()]
        else:
            multiband_linear = False
        if 'multiband_quadratic' in quadratic_matrix:
            multiband_quadratic = quadratic_matrix['multiband_quadratic'][()]
        else:
            multiband_quadratic = False

        # Get intersection between ifo and ROQ frequency samples. Required only for non-multibanded basis.
        if not (multiband_linear and multiband_quadratic):
            roq_idxs = {}
            ifo_idxs = {}
            for ifo in self.interferometers:
                if self.roq_params is not None:
                    self.perform_roq_params_check(ifo)
                    # Get scaled ROQ quantities
                    roq_scaled_minimum_frequency = self.roq_params['flow'] * self.roq_scale_factor
                    roq_scaled_maximum_frequency = self.roq_params['fhigh'] * self.roq_scale_factor
                    roq_scaled_segment_length = self.roq_params['seglen'] / self.roq_scale_factor
                    # Generate frequencies for the ROQ
                    roq_frequencies = create_frequency_series(
                        sampling_frequency=roq_scaled_maximum_frequency * 2,
                        duration=roq_scaled_segment_length)
                    roq_mask = roq_frequencies >= roq_scaled_minimum_frequency
                    roq_frequencies = roq_frequencies[roq_mask]
                    overlap_frequencies, ifo_idxs_this_ifo, roq_idxs_this_ifo = np.intersect1d(
                        ifo.frequency_array[ifo.frequency_mask], roq_frequencies,
                        return_indices=True)
                else:
                    overlap_frequencies = ifo.frequency_array[ifo.frequency_mask]
                    roq_idxs_this_ifo = np.arange(
                        linear_matrix['basis_linear'][str(idxs_in_prior_range['linear'][0])]['basis'].shape[1],
                        dtype=int)
                    ifo_idxs_this_ifo = np.arange(sum(ifo.frequency_mask))
                    if len(ifo_idxs_this_ifo) != len(roq_idxs_this_ifo):
                        raise ValueError(
                            "Mismatch between ROQ basis and frequency array for "
                            "{}".format(ifo.name))
                logger.info(
                    "Building ROQ weights for {} with {} frequencies between {} "
                    "and {}.".format(
                        ifo.name, len(overlap_frequencies),
                        min(overlap_frequencies), max(overlap_frequencies)))
                roq_idxs[ifo.name] = roq_idxs_this_ifo
                ifo_idxs[ifo.name] = ifo_idxs_this_ifo

        if multiband_linear:
            self._set_weights_linear_multiband(linear_matrix, idxs_in_prior_range['linear'])
        else:
            self._set_weights_linear(linear_matrix, idxs_in_prior_range['linear'], roq_idxs, ifo_idxs)

        if multiband_quadratic:
            self._set_weights_quadratic_multiband(quadratic_matrix, idxs_in_prior_range['quadratic'])
        else:
            self._set_weights_quadratic(quadratic_matrix, idxs_in_prior_range['quadratic'], roq_idxs, ifo_idxs)

    def _set_weights_linear(self, linear_matrix, basis_idxs, roq_idxs, ifo_idxs):
        """
        Setup the time-dependent linear ROQ weights. See https://dcc.ligo.org/LIGO-T2100125 for the detail of how to
        compute them.

        Parameters
        ==========
        linear_matrix : dictionary or h5py.File
            linear basis
        basis_idxs : array-like
            indexes of bases used for a run
        roq_idxs : dictionary
            dictionary whose keys are interferometer names and values are indexes of basis components intersecting
            frequency-domain data
        ifo_idxs : dictionary
            dictionary whose keys are interferometer names and values are indexes of frequency-domain data intersecting
            basis components

        """
        for ifo in self.interferometers:
            self.weights[ifo.name + '_linear'] = []
        time_space = self.weights['time_samples'][1] - self.weights['time_samples'][0]
        number_of_time_samples = int(self.interferometers.duration / time_space)
        start_idx = int(self.weights['time_samples'][0] / time_space)
        end_idx = int(self.weights['time_samples'][-1] / time_space)
        nonzero_idxs = {}
        data_over_psd = {}
        for ifo in self.interferometers:
            nonzero_idxs[ifo.name] = ifo_idxs[ifo.name] + int(
                ifo.frequency_array[ifo.frequency_mask][0] * self.interferometers.duration)
            data_over_psd[ifo.name] = ifo.frequency_domain_strain[ifo.frequency_mask][ifo_idxs[ifo.name]] / \
                ifo.power_spectral_density_array[ifo.frequency_mask][ifo_idxs[ifo.name]]
        try:
            import pyfftw
            ifft_input = pyfftw.empty_aligned(number_of_time_samples, dtype=complex)
            ifft_output = pyfftw.empty_aligned(number_of_time_samples, dtype=complex)
            ifft = pyfftw.FFTW(ifft_input, ifft_output, direction='FFTW_BACKWARD')
        except ImportError:
            pyfftw = None
            logger.warning("You do not have pyfftw installed, falling back to numpy.fft.")
            ifft_input = np.zeros(number_of_time_samples, dtype=complex)
            ifft = np.fft.ifft
        for basis_idx in basis_idxs:
            logger.info(f"Building linear ROQ weights for the {basis_idx}-th basis.")
            linear_matrix_single = linear_matrix['basis_linear'][str(basis_idx)]['basis']
            basis_size = linear_matrix_single.shape[0]
            for ifo in self.interferometers:
                ifft_input[:] *= 0.
                linear_weights = \
                    np.zeros((len(self.weights['time_samples']), basis_size), dtype=complex)
                for i in range(basis_size):
                    basis_element = linear_matrix_single[i][roq_idxs[ifo.name]]
                    ifft_input[nonzero_idxs[ifo.name]] = data_over_psd[ifo.name] * np.conj(basis_element)
                    linear_weights[:, i] = ifft(ifft_input)[start_idx:end_idx + 1]
                linear_weights *= 4. * number_of_time_samples / self.interferometers.duration
                self.weights[ifo.name + '_linear'].append(linear_weights)
        if pyfftw is not None:
            pyfftw.forget_wisdom()

    def _set_weights_linear_multiband(self, linear_matrix, basis_idxs):
        """
        Setup the time-dependent linear ROQ weights from multibanded basis

        Parameters
        ==========
        linear_matrix : dictionary or h5py.File
            linear basis
        basis_idxs : array-like
            indexes of bases used for a run

        """
        for ifo in self.interferometers:
            self.weights[ifo.name + '_linear'] = []
        Tbs = linear_matrix['durations_s_linear'][()] / self.roq_scale_factor
        start_end_frequency_bins = linear_matrix['start_end_frequency_bins_linear'][()]
        basis_dimension = np.sum(start_end_frequency_bins[:, 1] - start_end_frequency_bins[:, 0] + 1)
        fhigh_basis = np.max(start_end_frequency_bins[:, 1] / Tbs)
        # prepare time-shifted data, which is multiplied by basis
        tc_shifted_data = dict()
        for ifo in self.interferometers:
            over_whitened_frequency_data = np.zeros(int(fhigh_basis * ifo.duration) + 1, dtype=complex)
            over_whitened_frequency_data[np.arange(len(ifo.frequency_domain_strain))[ifo.frequency_mask]] = \
                ifo.frequency_domain_strain[ifo.frequency_mask] / ifo.power_spectral_density_array[ifo.frequency_mask]
            over_whitened_time_data = np.fft.irfft(over_whitened_frequency_data)
            tc_shifted_data[ifo.name] = np.zeros((basis_dimension, len(self.weights['time_samples'])), dtype=complex)
            start_idx_of_band = 0
            for b, Tb in enumerate(Tbs):
                start_frequency_bin, end_frequency_bin = start_end_frequency_bins[b]
                fs = np.arange(start_frequency_bin, end_frequency_bin + 1) / Tb
                Db = np.fft.rfft(
                    over_whitened_time_data[-int(2. * fhigh_basis * Tb):]
                )[start_frequency_bin:end_frequency_bin + 1]
                start_idx_of_next_band = start_idx_of_band + end_frequency_bin - start_frequency_bin + 1
                tc_shifted_data[ifo.name][start_idx_of_band:start_idx_of_next_band] = 4. / Tb * Db[:, None] * np.exp(
                    2. * np.pi * 1j * fs[:, None] * (self.weights['time_samples'][None, :] - ifo.duration + Tb))
                start_idx_of_band = start_idx_of_next_band
        # compute inner products
        for basis_idx in basis_idxs:
            logger.info(f"Building linear ROQ weights for the {basis_idx}-th basis.")
            linear_matrix_single = linear_matrix['basis_linear'][str(basis_idx)]['basis'][()]
            for ifo in self.interferometers:
                self.weights[ifo.name + '_linear'].append(
                    np.dot(np.conj(linear_matrix_single), tc_shifted_data[ifo.name]).T)

    def _set_weights_quadratic(self, quadratic_matrix, basis_idxs, roq_idxs, ifo_idxs):
        """
        Setup the quadratic ROQ weights

        Parameters
        ==========
        quadratic_matrix : dictionary or h5py.File
            quadratic basis
        basis_idxs : array-like
            indexes of bases used for a run
        roq_idxs : dictionary
            dictionary whose keys are interferometer names and values are indexes of basis components intersecting
            frequency-domain data
        ifo_idxs : dictionary
            dictionary whose keys are interferometer names and values are indexes of frequency-domain data intersecting
            basis components

        """
        for ifo in self.interferometers:
            self.weights[ifo.name + '_quadratic'] = []
        for basis_idx in basis_idxs:
            logger.info(f"Building quadratic ROQ weights for the {basis_idx}-th basis.")
            quadratic_matrix_single = quadratic_matrix['basis_quadratic'][str(basis_idx)]['basis'][()].real
            for ifo in self.interferometers:
                self.weights[ifo.name + '_quadratic'].append(
                    4. / ifo.strain_data.duration * np.dot(
                        quadratic_matrix_single[:, roq_idxs[ifo.name]],
                        1 / ifo.power_spectral_density_array[ifo.frequency_mask][ifo_idxs[ifo.name]]))
            del quadratic_matrix_single

    def _set_weights_quadratic_multiband(self, quadratic_matrix, basis_idxs):
        """
        Setup the quadratic ROQ weights from multibanded basis

        Parameters
        ==========
        quadratic_matrix : dictionary or h5py.File
            quadratic basis
        basis_idxs : array-like
            indexes of bases used for a run

        """
        for ifo in self.interferometers:
            self.weights[ifo.name + '_quadratic'] = []
        Tbs = quadratic_matrix['durations_s_quadratic'][()] / self.roq_scale_factor
        start_end_frequency_bins = quadratic_matrix['start_end_frequency_bins_quadratic'][()]
        basis_dimension = np.sum(start_end_frequency_bins[:, 1] - start_end_frequency_bins[:, 0] + 1)
        fhigh_basis = np.max(start_end_frequency_bins[:, 1] / Tbs)
        # prepare coefficients multiplied by basis
        multibanded_inverse_psd = dict()
        for ifo in self.interferometers:
            inverse_psd_frequency = np.zeros(int(fhigh_basis * ifo.duration) + 1)
            inverse_psd_frequency[np.arange(len(ifo.power_spectral_density_array))[ifo.frequency_mask]] = \
                1. / ifo.power_spectral_density_array[ifo.frequency_mask]
            inverse_psd_time = np.fft.irfft(inverse_psd_frequency)
            multibanded_inverse_psd[ifo.name] = np.zeros(basis_dimension)
            start_idx_of_band = 0
            for b, Tb in enumerate(Tbs):
                start_frequency_bin, end_frequency_bin = start_end_frequency_bins[b]
                number_of_samples_half = int(fhigh_basis * Tb)
                start_idx_of_next_band = start_idx_of_band + end_frequency_bin - start_frequency_bin + 1
                multibanded_inverse_psd[ifo.name][start_idx_of_band:start_idx_of_next_band] = 4. / Tb * np.fft.rfft(
                    np.append(inverse_psd_time[:number_of_samples_half], inverse_psd_time[-number_of_samples_half:])
                )[start_frequency_bin:end_frequency_bin + 1].real
                start_idx_of_band = start_idx_of_next_band
        # compute inner products
        for basis_idx in basis_idxs:
            logger.info(f"Building quadratic ROQ weights for the {basis_idx}-th basis.")
            quadratic_matrix_single = quadratic_matrix['basis_quadratic'][str(basis_idx)]['basis'][()].real
            for ifo in self.interferometers:
                self.weights[ifo.name + '_quadratic'].append(
                    np.dot(quadratic_matrix_single, multibanded_inverse_psd[ifo.name]))

    def save_weights(self, filename, format='npz'):
        """
        Save ROQ weights into a single file. format should be json, npz, or hdf5. For weights from multiple bases, hdf5
        is only the possible option.

        Parameters
        ==========
        filename : str
        format : str

        """
        if format not in ['json', 'npz', 'hdf5']:
            raise IOError(f"Format {format} not recognized.")
        if format not in filename:
            filename += "." + format
        logger.info(f"Saving ROQ weights to {filename}")
        if format == 'json' or format == 'npz':
            if self.number_of_bases_linear > 1 or self.number_of_bases_quadratic > 1:
                raise ValueError(f'Format {format} not compatible with multiple bases')
            weights = dict()
            weights['time_samples'] = self.weights['time_samples']
            for basis_type in ['linear', 'quadratic']:
                for ifo in self.interferometers:
                    key = f'{ifo.name}_{basis_type}'
                    weights[key] = self.weights[key][0]
            if format == 'json':
                with open(filename, 'w') as file:
                    json.dump(weights, file, indent=2, cls=BilbyJsonEncoder)
            else:
                np.savez(filename, **weights)
        else:
            import h5py
            with h5py.File(filename, 'w') as f:
                f.create_dataset('time_samples',
                                 data=self.weights['time_samples'])
                for basis_type in ['linear', 'quadratic']:
                    key = f'prior_range_{basis_type}'
                    if key in self.weights:
                        grp = f.create_group(key)
                        for param_name in self.weights[key]:
                            grp.create_dataset(
                                param_name, data=self.weights[key][param_name])
                    key = f'frequency_nodes_{basis_type}'
                    if key in self.weights:
                        grp = f.create_group(key)
                        for i in range(len(self.weights[key])):
                            grp.create_dataset(
                                str(i), data=self.weights[key][i])
                    for ifo in self.interferometers:
                        key = f"{ifo.name}_{basis_type}"
                        grp = f.create_group(key)
                        for i in range(len(self.weights[key])):
                            grp.create_dataset(
                                str(i), data=self.weights[key][i])

    def load_weights(self, filename, format=None):
        """
        Load ROQ weights. format should be json, npz, or hdf5. json or npz file is assumed to contain weights from a
        single basis

        Parameters
        ==========
        filename : str
        format : str

        """
        if format is None:
            format = filename.split(".")[-1]
        if format not in ["json", "npz", "hdf5"]:
            raise IOError(f"Format {format} not recognized.")
        logger.info(f"Loading ROQ weights from {filename}")
        if format == "json" or format == "npz":
            # Old file format assumed to contain only a single basis
            if format == "json":
                with open(filename, 'r') as file:
                    weights = json.load(file, object_hook=decode_bilby_json)
            else:
                # Wrap in dict to load data into memory
                weights = dict(np.load(filename))
            for basis_type in ['linear', 'quadratic']:
                for ifo in self.interferometers:
                    key = f'{ifo.name}_{basis_type}'
                    weights[key] = [weights[key]]
        else:
            weights = dict()
            import h5py
            with h5py.File(filename, 'r') as f:
                weights['time_samples'] = f['time_samples'][()]
                for basis_type in ['linear', 'quadratic']:
                    key = f'prior_range_{basis_type}'
                    if key in f:
                        idxs_in_prior_range, selected_prior_ranges = \
                            self._select_prior_ranges(f[key])
                        weights[key] = selected_prior_ranges
                    else:
                        idxs_in_prior_range = [0]
                    key = f'frequency_nodes_{basis_type}'
                    if key in f:
                        weights[key] = [f[key][str(i)][()]
                                        for i in idxs_in_prior_range]
                    for ifo in self.interferometers:
                        key = f"{ifo.name}_{basis_type}"
                        weights[key] = [f[key][str(i)][()]
                                        for i in idxs_in_prior_range]
        return weights

    def _get_time_resolution(self):
        """
        This method estimates the time resolution given the optimal SNR of the
        signal in the detector. This is then used when constructing the weights
        for the ROQ.

        A minimum resolution is set by assuming the SNR in each detector is at
        least 10. When the SNR is not available the SNR is assumed to be 30 in
        each detector.

        Returns
        =======
        delta_t: float
            Time resolution
        """

        def calc_fhigh(freq, psd, scaling=20.):
            """

            Parameters
            ==========
            freq: array-like
                Frequency array
            psd: array-like
                Power spectral density
            scaling: float
                SNR dependent scaling factor

            Returns
            =======
            f_high: float
                The maximum frequency which must be considered
            """
            from scipy.integrate import simps
            integrand1 = np.power(freq, -7. / 3) / psd
            integral1 = simps(integrand1, freq)
            integrand3 = np.power(freq, 2. / 3.) / (psd * integral1)
            f_3_bar = simps(integrand3, freq)

            f_high = scaling * f_3_bar**(1 / 3)

            return f_high

        def c_f_scaling(snr):
            return (np.pi**2 * snr**2 / 6)**(1 / 3)

        inj_snr_sq = 0
        for ifo in self.interferometers:
            inj_snr_sq += max(10, ifo.meta_data.get('optimal_SNR', 30))**2

        psd = ifo.power_spectral_density_array[ifo.frequency_mask]
        freq = ifo.frequency_array[ifo.frequency_mask]
        fhigh = calc_fhigh(freq, psd, scaling=c_f_scaling(inj_snr_sq**0.5))

        delta_t = fhigh**-1

        # Apply a safety factor to ensure the time step is short enough
        delta_t = delta_t / 5

        # duration / delta_t needs to be a power of 2 for IFFT
        number_of_time_samples = max(
            self.interferometers.duration / delta_t,
            self.interferometers.frequency_array[-1] * self.interferometers.duration + 1)
        number_of_time_samples = int(2**np.ceil(np.log2(number_of_time_samples)))
        delta_t = self.interferometers.duration / number_of_time_samples
        logger.info("ROQ time-step = {}".format(delta_t))
        return delta_t

    def _rescale_signal(self, signal, new_distance):
        for kind in ['linear', 'quadratic']:
            for mode in signal[kind]:
                signal[kind][mode] *= self._ref_dist / new_distance


class BilbyROQParamsRangeError(Exception):
    pass



import copy
import os
import shutil
import unittest
from unittest.mock import MagicMock
from parameterized import parameterized

import numpy as np

import bilby
from bilby.core import prior


class TestSampler(unittest.TestCase):
    def setUp(self):
        likelihood = bilby.core.likelihood.Likelihood()
        likelihood.parameters = dict(a=1, b=2, c=3)
        delta_prior = prior.DeltaFunction(peak=0)
        delta_prior.rescale = MagicMock(return_value=prior.DeltaFunction(peak=1))
        delta_prior.prob = MagicMock(return_value=1)
        delta_prior.sample = MagicMock(return_value=0)
        uniform_prior = prior.Uniform(0, 1)
        uniform_prior.rescale = MagicMock(return_value=prior.Uniform(0, 2))
        uniform_prior.prob = MagicMock(return_value=1)
        uniform_prior.sample = MagicMock(return_value=0.5)

        priors = dict(a=delta_prior, b="string", c=uniform_prior)
        likelihood.log_likelihood_ratio = MagicMock(return_value=1)
        likelihood.log_likelihood = MagicMock(return_value=2)
        test_directory = "test_directory"
        if os.path.isdir(test_directory):
            os.rmdir(test_directory)
        self.sampler = bilby.core.sampler.Sampler(
            likelihood=likelihood,
            priors=priors,
            outdir=test_directory,
            use_ratio=False,
            skip_import_verification=True,
        )

    def tearDown(self):
        del self.sampler

    def test_search_parameter_keys(self):
        expected_search_parameter_keys = ["c"]
        self.assertListEqual(
            self.sampler.search_parameter_keys, expected_search_parameter_keys
        )

    def test_fixed_parameter_keys(self):
        expected_fixed_parameter_keys = ["a"]
        self.assertListEqual(
            self.sampler.fixed_parameter_keys, expected_fixed_parameter_keys
        )

    def test_ndim(self):
        self.assertEqual(self.sampler.ndim, 1)

    def test_kwargs(self):
        self.assertDictEqual(self.sampler.kwargs, {})

    def test_label(self):
        self.assertEqual(self.sampler.label, "label")

    def test_prior_transform_transforms_search_parameter_keys(self):
        self.sampler.prior_transform([0])
        expected_prior = prior.Uniform(0, 1)
        self.assertListEqual(
            [self.sampler.priors["c"].minimum, self.sampler.priors["c"].maximum],
            [expected_prior.minimum, expected_prior.maximum],
        )

    def test_prior_transform_does_not_transform_fixed_parameter_keys(self):
        self.sampler.prior_transform([0])
        self.assertEqual(
            self.sampler.priors["a"].peak, prior.DeltaFunction(peak=0).peak
        )

    def test_log_prior(self):
        self.assertEqual(self.sampler.log_prior({1}), 0.0)

    def test_log_likelihood_with_use_ratio(self):
        self.sampler.use_ratio = True
        self.assertEqual(self.sampler.log_likelihood([0]), 1)

    def test_log_likelihood_without_use_ratio(self):
        self.sampler.use_ratio = False
        self.assertEqual(self.sampler.log_likelihood([0]), 2)

    def test_log_likelihood_correctly_sets_parameters(self):
        expected_dict = dict(a=0, b=2, c=0)
        _ = self.sampler.log_likelihood([0])
        self.assertDictEqual(self.sampler.likelihood.parameters, expected_dict)

    def test_get_random_draw(self):
        self.assertEqual(self.sampler.get_random_draw_from_prior(), np.array([0.5]))

    def test_base_run_sampler(self):
        sampler_copy = copy.copy(self.sampler)
        self.sampler.run_sampler()
        self.assertDictEqual(sampler_copy.__dict__, self.sampler.__dict__)

    def test_bad_value_nan(self):
        self.sampler._check_bad_value(val=np.nan, warning=False, theta=None, label=None)

    def test_bad_value_np_abs_nan(self):
        self.sampler._check_bad_value(
            val=np.abs(np.nan), warning=False, theta=None, label=None
        )

    def test_bad_value_abs_nan(self):
        self.sampler._check_bad_value(
            val=abs(np.nan), warning=False, theta=None, label=None
        )

    def test_bad_value_pos_inf(self):
        self.sampler._check_bad_value(val=np.inf, warning=False, theta=None, label=None)

    def test_bad_value_neg_inf(self):
        self.sampler._check_bad_value(
            val=-np.inf, warning=False, theta=None, label=None
        )

    def test_bad_value_pos_inf_nan_to_num(self):
        self.sampler._check_bad_value(
            val=np.nan_to_num(np.inf), warning=False, theta=None, label=None
        )

    def test_bad_value_neg_inf_nan_to_num(self):
        self.sampler._check_bad_value(
            val=np.nan_to_num(-np.inf), warning=False, theta=None, label=None
        )


samplers = [
    "bilby_mcmc",
    "dynamic_dynesty",
    "dynesty",
    "emcee",
    "kombine",
    "ptemcee",
    "zeus",
]


class GenericSamplerTest(unittest.TestCase):
    def setUp(self):
        self.likelihood = bilby.core.likelihood.Likelihood(dict())
        self.priors = bilby.core.prior.PriorDict(
            dict(a=bilby.core.prior.Uniform(0, 1), b=bilby.core.prior.Uniform(0, 1))
        )

    def tearDown(self):
        if os.path.isdir("outdir"):
            shutil.rmtree("outdir")

    @parameterized.expand(samplers)
    def test_pool_creates_properly_no_pool(self, sampler_name):
        sampler = bilby.core.sampler.IMPLEMENTED_SAMPLERS[sampler_name](
            self.likelihood, self.priors
        )
        sampler._setup_pool()
        if sampler_name == "kombine":
            from kombine import SerialPool

            self.assertIsInstance(sampler.pool, SerialPool)
            pass
        else:
            self.assertIsNone(sampler.pool)

    @parameterized.expand(samplers)
    def test_pool_creates_properly_pool(self, sampler):
        sampler = bilby.core.sampler.IMPLEMENTED_SAMPLERS[sampler](
            self.likelihood, self.priors, npool=2
        )
        sampler._setup_pool()
        if hasattr(sampler, "setup_sampler"):
            sampler.setup_sampler()
        self.assertEqual(sampler.pool._processes, 2)
        sampler._close_pool()


class ReorderLikelihoodsTest(unittest.TestCase):
    def setUp(self):
        self.unsorted_ln_likelihoods = np.array([1, 5, 2, 5, 1])
        self.unsorted_samples = np.array([[0, 1], [1, 1], [1, 0], [0, 0], [0, 1]])
        self.sorted_samples = np.array([[0, 1], [0, 1], [1, 0], [1, 1], [0, 0]])
        self.sorted_ln_likelihoods = np.array([1, 1, 2, 5, 5])

    def tearDown(self):
        pass

    def test_ordering(self):
        func = bilby.core.sampler.base_sampler.NestedSampler.reorder_loglikelihoods
        sorted_ln_likelihoods = func(
            self.unsorted_ln_likelihoods, self.unsorted_samples, self.sorted_samples
        )
        self.assertTrue(
            np.array_equal(sorted_ln_likelihoods, self.sorted_ln_likelihoods)
        )


if __name__ == "__main__":
    unittest.main()
